from contextlib import contextmanager
import logging
import math
import pathlib
import subprocess
import re
from typing import (
    Any,
    Callable,
    List,
    Optional,
    Tuple,
    TypeAlias,
    Union,
    Dict,
    Set,
)
import numpy as np
import torch
import torch.amp
import torch.nn as nn
import torchinfo
from . import _logging

"""
This file aims to be a project-independent training-loop and associated
scaffolding.
"""

_logger = logging.getLogger(__name__)

"""
Max training time before a new recovery is made.

Standard checkpointing only happens every epoch. But a checkpoint for recovery
purposes will be made every 30 minutes. Only a single recovery is kept.
"""
RECOVERY_CKPT_PERIOD_SEC = 30 * 60

"""
Here we take part in the rite of of passage for a deep learning project by
yet again reinventing the training loop architecture. No one wants their 
project stitched together with the callbacks of some soon to be abandonded or 
rewritten DL framework.
"""


def to_device(x: Union[list, tuple, dict, torch.Tensor], device):
    """
    Recursively move all tensors in x to the given device.

    Supports lists, tuples, dictionaries, and torch.Tensor.
    """
    if isinstance(x, torch.Tensor):
        return x.to(device)
    elif isinstance(x, dict):
        return {k: to_device(v, device) for k, v in x.items()}
    elif isinstance(x, (list, tuple)):
        return [to_device(v, device) for v in x]
    raise ValueError(f"Unsupported type: {type(x)}")


def gpus_ordered_by_mem_used(exclude: Optional[Set[int]] = None):
    """
    Return a list of GPU indices ordered by memory used (increasing).

    You probably want to use the first GPU in the returned list.
    """
    try:
        out_str = subprocess.check_output(
            # -A<X> prints X lines after the match.
            # Full output is like, one line per GPU:
            #  Used        : 11MiB
            #  Used        : 66MiB
            "nvidia-smi -q -d Memory | grep -A4 GPU | grep Used",
            shell=True,
        )
    except subprocess.CalledProcessError as e:
        _logger.error(
            "Failed to parse nvidia-smi output looking for GPU information. "
            "Most likely, nvidia-smi has changed it's output format, and this "
            "code needs a simple tweak."
        )
        raise e
    mems = re.findall(r"Used\s+:\s+(\d+)\s+MiB", out_str.decode("utf-8"))
    if not len(mems):
        raise ValueError(f"Unexpected output: {out_str}")
    res = sorted(range(len(mems)), key=lambda i: int(mems[i]))
    if exclude is not None:
        res = [i for i in res if i not in exclude]
    return res


class ModelException(Exception):
    """Raised when a model causes training to fail.

    This allows callers to catch this exception and move on to the next model.
    """

    pass


class Trainable:
    """Encapsulates a dataset, model input-output and loss function.

    This class is needed in order to be able to train multiple models and
    configurations with the same training function. The training function
    is too general to know about how to route the data into and out of a model,
    evaluate the model or how to take a model output and create a prediction.

    Redesign from function parameters to a class
    --------------------------------------------
    The class began as a dumb grouping of the parameters to the train function:
    train_ds, val_ds, test_ds, model, loss_fn, forward_fn, val_fn, and more—there
    were so many parameters that they were grouped together into a NamedTuple.
    However, functions like forward_fn and val_fn would need to be given
    the model, loss_fn and forward_fn in order to operate. This is the exact
    encapsulation behaviour that classes achieve, so the NamedTuple was made
    into a class. Leaning into the use of classes, forward_fn and val_fn were
    made into methods of the class, while the rest became properties. This
    change is noteworthy as customizing the forward or validation functions
    now requires defining a new class, rather than simply passing in a new
    function. Although, nothing is stopping someone from defining a class that
    simply wraps a function and passes the arguments through.

    Flexibility to vary the data format
    -----------------------------------
    Why things belongs inside or outside this class can be understood by
    realizing that the nature of the datasets are known here. As such,
    any function that needs to extract the individual parts of a dataset
    sample will need to know what is in each sample. Such a function is a
    good candidate to appear in this class.

    While in some cases you can separate the datasets from the models, this
    isn't always easy or a good idea. A model for ImageNet can easily be
    separated from the dataset, as the inputs and outputs are so standard; but
    for the spike prediction, the model output is quite variable. Consider the
    distance array model which outputs an array, whereas a Poisson-distribution
    model will output a single number. Training is done with these outputs, and
    involves the dataset producing sample tuples that have appropriate elements
    (distance fields, for example). The actual inference is an additional
    calculation using these outputs.

    In other words, the procedure of taking the model input and model output
    from a dataset sample and feeding it to the model, then calculating the
    loss and doing the inference—none of these steps can be abstracted to be
    ignorant of either the model or the dataset.

    Handles dataloader creation
    ---------------------------
    train.py handled this responsibility for a number of months; however,
    it became a problem once I needed more control over the evaluation
    function. For evaluation, I wanted to run the slow inference proceedure
    on a subset of the cells and additionally create input-output videos
    for them. This was specific to just one of the trainables. The ideal
    approach would be for the evaluation function to get or create a number
    of extra data loaders (one for each cluster) and then run the more detailed
    evaluation on each. There wasn't a good way to do this without moving
    the full responsibility of dataloader handing into the Trainable class.
    The current approach is to use an intermediate (DatasetManager) that lazily
    creates datasets; this allows trainables to create multiple variations of
    a dataset as needed, without them being created up front.

    Other libraries
    ---------------
    Compared to Keras and FastAI: Trainable encapsulates a lot less than
    Keras's Model or FastAI's Learner.

    At this point, I'm not eager to use the FastAI API, as I don't
    want to discover later that it's too limiting in some certain way. It's
    quite possible that it's already too prescriptive. Reading the docs, it's
    not clear what parts of Learner's internals are exposed for customization.
    If all "_" prefixed methods are not meant to be customized, then it's
    already too restrictive. Notably, there seems to be an expected format for
    the elements of the dataset, which I want to avoid. The reason for this is
    that the distance arrays are intermediate results, and while I want to
    train on them, I would like to evaluate based on quick approximate
    inference and make predictions using much more expensive and accurate
    inference routines. So the data doesn't fall nicely into (X,y) type data,
    and the metrics are not consistent across training and evaluation.

    In addition, at least at the momemt, FastAI's library provides a lot more
    abstraction/generalization than I need, which can make it harder for
    myself (or others) to understand what is going on. This might end up being
    a mistake, as the growing code might reveal itself to provide abstraction
    boundaries that are already handled nicely in FastAI.
    """

    def __init__(self, model: torch.nn.Module, label: str):
        """
        Args:
            model: the PyTorch model to train.
            label: a string label for this trainable.
        """
        self.model = model
        self.label = label

    def train_ds(self):
        raise NotImplementedError("Override")

    def val_ds(self):
        raise NotImplementedError("Override")

    def test_ds(self):
        raise NotImplementedError("Override")

    def forward(self, sample):
        """Run the model forward.

        Args:
            sample: a single draw from the train or validation data loader.

        Returns:
            (output, loss): the model output and the loss, as a tuple.
        """
        raise NotImplementedError("Override")

    """
    Some considerations about the evaluate functions.
       - The evaluation functions should be given the power and responsibility
         to create the datasets. This is because they may want to change the
         nature of a dataset. For example, the spike prediction models want
         to change ds stride and to filter in/out certain clusters.
       - The evaluation functions should not need to care about how to create
         a dataloader from a dataset. Things like worker counts and batch
         sizes should not need to be handled.
       - There should two separate routines: one for the training dataset and
         one for the validation dataset. The reason for the distinction 
         will mostly often be the sizes of the datasets. 
       - It might be desirable to have a third function that will be run
         at the end of training, or outside of training, and is expected to be 
         a long-running. 
    """

    def evaluate_train(self, dl_fn):
        """Run the evaluation procedure on the train dataset.

        Args:
            dl_fn: a function to convert the train dataset into a dataloader.

        Returns:
            metrics: a str:float dictionary containing evaluation metrics. It
                is expected that this dictionary at least contains 'loss' and
                'accuracy' metrics.
        """
        raise NotImplementedError("Override")

    def evaluate_val(self, dl_fn):
        """Run the evaluation procedure on the val dataset.

        Args:
            dl_fn: a function to convert the train dataset into a dataloader.

        Returns:
            metrics: a str:float dictionary containing evaluation metrics. It
                is expected that this dictionary at least contains 'loss' and
                'accuracy' metrics.
        """
        raise NotImplementedError("Override")

    def evaluate_full(self, ds, dl_fn):
        """Run a pontentially long evaluation procedure on the given dataset.

        This isn't implemented anywhere yet. The idea here is to carve out
        some space for running a longer evaluation procedure on a dataset. This
        function is not called during the training routine.

        Args:
            ds: the dataset to evaluate on.
            dl_fn: a function to convert the train dataset into a dataloader.
        """
        raise NotImplementedError("Override")

    def in_device(self):
        """Returns the device on which the model expects input to be located.

        Most likely, the whole model is on a single device, and it is
        sufficient to use `next(self.model.parameters()).device`.
        """
        raise NotImplementedError("Override")

    def __str__(self) -> str:
        return f"Trainable ({self.label})"

    def model_summary(self, batch_size: int) -> str:
        """Returns a detailed description of the model.

        Args:
            batch_size: the batch size to use when creating the model summary.

        At the moment, this is called by train() with the intent of saving
        out a file containing info like torchinfo.summary. The torch module
        input shape isn't known by train(), so the actual summary creation
        must be done somewhere like Trainable.

        Override this to add more features.
        """
        return f"Trainable ({self.label})"


class DatasetManager:
    """
    Create train, val & test Pytorch Datasets.

    The train, val & test datasets must be created almost identically, except
    for small changes to options like shuffle.

    The reason for choosing factory functions to be passed around instead of
    the actual datasets is that there are cases where we need to create
    more, and the initial train, val, test datasets are not sufficient.
    Alternatively, we want access to dataloaders and not datasets or vise-versa
    The case that initiated this refactor was the need to split a validation
    set into multiple 1-cluster datasets for the purpose of evaluating
    autoregressive inference.

    Another benefit is that the datasets aren't created eagerly, which can
    save memory if they aren't needed, which is typically the case for the
    test dataset while training.

    train.py will construct the DatasetManager and give it to the Trainable.
    The DatasetManager will be held by the Trainable, and so train.py will
    no longer need to construct and pass around dataloaders.
    """

    def train_ds(self) -> torch.utils.data.Dataset:
        raise NotImplementedError()

    def val_ds(self) -> torch.utils.data.Dataset:
        raise NotImplementedError()

    def test_ds(self) -> torch.utils.data.Dataset:
        raise NotImplementedError()


class BasicDatasetManager(DatasetManager):
    """Store and return the datasets as-is.

    Useful for simple cases. To handle stastistics of the training set, like the
    mean and standard deviation, or whatever else is needed, use the
    train_ds_attrs, which will convert the key-value entries to class
    properties. just make sure to use valid python identifiers for the keys.

    Improvement: we could allow train_ds, val_ds and test_ds to optionally be
    functions, which would allow the datasets to be created lazily and fresh
    each time.
    """

    def __init__(self, train_ds, val_ds, test_ds, train_ds_attrs=None):
        """
        Args:
            train_ds: the training dataset.
            val_ds: the validation dataset.
            test_ds: the test dataset.
            train_ds_attrs: a dictionary of attributes to store with the
                training dataset. These attributes can be accessed as properties
                of the DatasetManager. Use this for values like mean, std and
                so on—values used for things like normalization.

        """
        self._train_ds = train_ds
        self._val_ds = val_ds
        self._test_ds = test_ds
        # Store the attributes dictionary
        self._train_ds_attrs = train_ds_attrs or {}
        # Dynamically create properties for each attribute
        for attr_name in self._train_ds_attrs.keys():
            if not attr_name.isidentifier():
                raise ValueError(f"Invalid attribute name: {attr_name}")

            # Create a property getter for this attribute
            # The following will reference the final value of the loop variable
            # attr_name, which would be a bug.
            # def getter(self):
            #     return self._train_ds_attrs[attr_name]
            # Instead we do:
            def getter(self, attr_name=attr_name):
                return self._train_ds_attrs[attr_name]

            setattr(self.__class__, attr_name, property(getter))

    def train_ds(self) -> torch.utils.data.Dataset:
        return self._train_ds

    def val_ds(self) -> torch.utils.data.Dataset:
        return self._val_ds

    def test_ds(self) -> torch.utils.data.Dataset:
        return self._test_ds


class BaseTrainable(Trainable):
    """
    Covers basic Trainable functionality.
    """

    def __init__(
        self,
        ds_mgr: DatasetManager,
        model: torch.nn.Module,
        label: str,
        # For speed, you probably want to cache.
        cache_dl: bool = True,
    ):
        super().__init__(model, label)
        self.ds_mgr = ds_mgr
        self._train_ds = None
        self._val_ds = None
        self._test_ds = None
        self.collate_fn = None
        self.cache_dl = cache_dl
        # For evaluation only:
        self._train_dl = None
        self._val_dl = None

    def train_ds(self) -> torch.utils.data.Dataset:
        if self._train_ds is None:
            self._train_ds = self.ds_mgr.train_ds()
        return self._train_ds

    def val_ds(self) -> torch.utils.data.Dataset:
        if self._val_ds is None:
            self._val_ds = self.ds_mgr.val_ds()
        return self._val_ds

    def test_ds(self) -> torch.utils.data.Dataset:
        if self._test_ds is None:
            self._test_ds = self.ds_mgr.test_ds()
        return self._test_ds

    def train_dl(self, dl_fn):
        if self._train_dl is None or not self.cache_dl:
            self._train_dl = dl_fn(self.train_ds(), collate_fn=self.collate_fn)
        return self._train_dl

    def val_dl(self, dl_fn):
        if self._val_dl is None or not self.cache_dl:
            self._val_dl = dl_fn(self.val_ds(), collate_fn=self.collate_fn)
        return self._val_dl

    def in_device(self):
        return next(self.model.parameters()).device

    def __str__(self) -> str:
        return f"Trainable ({self.label})"

    def model_summary(self, batch_size: int) -> str:
        dl = torch.utils.data.DataLoader(self.train_ds(), batch_size=batch_size)
        sample = next(iter(dl))
        X, mask, y = sample
        X = X.to(self.in_device())
        mask = mask.to(self.in_device())
        res = torchinfo.summary(
            self.model,
            input_data=(X, mask),
            col_names=["input_size", "output_size", "mult_adds", "num_params"],
            device=self.in_device(),
            depth=4,
        )
        return res

    # Commented out so that sub classes are free to override evaluate_train and
    # evaluate_val and change signature of evaluate().
    # def evaluate(self, dl):
    #     raise NotImplementedError("Override")

    def evaluate_train(self, dl_fn):
        return self.evaluate(self.train_dl(dl_fn))

    def evaluate_val(self, dl_fn):
        # We don't manage the model train/val state here, as it could be a
        #
        if self.model.training:
            _logger.warning("Model should be in eval mode")
        return self.evaluate(self.val_dl(dl_fn))


def _create_dataloaders(
    batch_size,
    eval_batch_size,
    n_workers,
    pin_memory,
    persistent_workers,
    samples_per_epoch=None,
):
    """
    Create dataloaders for train & val datasets.

    What is returned is a pair of factory functions λ.ds → dl.

    The test dataset is not included because it is not used during training.
    So far, the test_ds has never been needed during training, so it has
    not been included here.

    History note: it use to be the case that the dataloaders were created
    here and returned. However, this was changed for a number of reasons.
    Firstly, having dataset creation be lazy saves memory, and so dataloader
    creation should also be lazy. Secondly, we want to give more control to
    the Trainable, and instead of passing in a dataloader, we want to just
    say "please evaluate using the train dataset". We don't need to give in
    the dataset, as the Trainable is where train.py got access to it anyway.
    This might suggest a function like `trainable.evaluate_train(self)`;
    however, one missing piece is that the Trainable doesn't know how to
    create a dataloader. A few options remain:
        1. We give this responsibility to the DatasetManager. The downside of
        this is that the DatasetManager starts becoming a bit omniscient, and
        creating one becomes a bit of an unnecessary pain when we don't intend
        to train and may only want to run some inference.
        2. We could do 1. above, but leave a lot of default initialized
        options that we would expect train.py to set, such as batch_size.
        3. Pass a factory function λ.ds → dl to the Trainable when we call
        `trainable.evaluate_train(self)`. This is the approach taken here. We
        leave train.py with the responsibility of creating the dataloaders,
        and we keep the DatasetManager as a simple factory for datasets.
        A huge issue with 3 is that we can't reuse dataloaders! Adds massive
        time to evaluation. I'm starting to think the DatasetManager should
        do this.
        4. We could pass both a dataset and a dataloader or a dataset and a
        factory function λ.ds → d. This wont work, as in many cases the creation
        of the dataset must be done by the evaluate() function. For example,
        setting the stride of the dataset.
    """
    if persistent_workers:
        raise NotImplementedError(
            "persistent_workers is not yet supported due"
            "to segfault issue and dataloader exceeding dataset limits."
        )

    # Setting pin_memory=True. This is generally recommended when training on
    # Nvidia GPUs. See:
    #   - https://discuss.pytorch.org/t/when-to-set-pin-memory-to-true/19723
    #   - https://developer.nvidia.com/blog/how-optimize-data-transfers-cuda-cc/
    def sampler_fn(ds):
        if samples_per_epoch is None:
            return None
        # The comments suggest that replacement=False and num_samples can't be
        # set at the same time; however, inspecting the code suggests it's fine.
        # https://pytorch.org/docs/stable/_modules/torch/utils/data/sampler.html#Sampler
        sampler = torch.utils.data.RandomSampler(
            ds, replacement=False, num_samples=samples_per_epoch
        )
        return sampler

    def train_dl_fn(ds, collate_fn=None):
        res = torch.utils.data.DataLoader(
            ds,
            batch_size=batch_size,
            # shuffle must be False if a custom sampler is used.
            shuffle=False if samples_per_epoch is not None else True,
            drop_last=False,
            num_workers=n_workers,
            pin_memory=pin_memory,
            # Possible segfault issues when enabled. Also, possible issue with
            # dataloader passing beyond the epoch limit.
            # Disabled until fixed
            persistent_workers=persistent_workers,
            sampler=sampler_fn(ds),
            collate_fn=collate_fn,
        )
        return res

    def val_dl_fn(ds, collate_fn=None):
        res = torch.utils.data.DataLoader(
            ds,
            batch_size=eval_batch_size,
            # For debugging, it's nice to see a variety:
            shuffle=True,
            drop_last=False,
            num_workers=n_workers,
            pin_memory=pin_memory,
            # Temp disabled until seg fault is fixed.
            persistent_workers=persistent_workers,
            collate_fn=collate_fn,
        )
        return res

    return train_dl_fn, val_dl_fn


@contextmanager
def evaluating(model):
    """
    Context manager to set the model to eval mode and then back to train mode.

    Used this to prevent an exception leading to unexpected training state.
    """
    original_mode = model.training
    model.eval()
    try:
        model.eval()
        # TODO: torch compile forces model.training => True.
        # assert not model.training
        yield
    finally:
        # Switch back to the original training mode.
        model.train(original_mode)
        assert model.training == original_mode


class TrainCallbackList:
    """
    Document the available training callbacks.
    """

    def on_batch_start(self, step: int):
        pass

    def on_batch_end(self, step: int):
        pass

    def on_eval_start(self, step: int):
        pass

    def on_eval_end(self, step: int):
        pass


class Callback:
    """
    Currently, training is done in a stateless manner in train().

    An alternative is to have a stateful training setup, by creating a
    class Trainer that keeps data such as step and optimizer as properties.
    The benefit of a stateful approach is that callbacks can have a reference
    to the trainer and all callbacks can have an empty signature. The downside
    is that the callbacks are now coupled to the trainer, and so it's harder
    to reuse them in other contexts. It's clearly far more powerful to have
    access to a Trainer, and I'm leaning towards this approach. For now,
    we will be satisfied with a very basic callback interface without access
    to a stateful trainer.

    Another reason for a stateful approach is to allow "Control" callbacks:
    callbacks that can change the training state. Useful training state changes
    include changing the current learning rate, increasing/decreasing the
    number of epochs, early stopping, perturbing the model. A callback could
    hold a Queue of Actions which is populated by another thread, such as
    a gui.

    A compromise is no not have a Trainer, but have a TrainState object that can
    be edited by callbacks and is used by the train() function.
    """

    def before_train(self, trainable: Trainable, step: int):
        """
        Called after train() has completed setup, but before the first batch.

        Setup includes things like creating any loggers and configuring
        the optimizer.

        Step is 0 unless training is being resumed. Although, there currently
        isn't any support for resuming with an initialized step.
        """
        pass

    def before_batch(self, trainable: Trainable, step: int):
        pass

    def after_batch(self, trainable: Trainable, step: int):
        pass

    def after_train(self, trainable: Trainable, step: int):
        pass

    def before_eval(self, trainable: Trainable, step: int):
        pass

    def after_eval(self, trainable: Trainable, step: int):
        pass


class TensorTag(torch.nn.Module):
    """Use this module to log tensors.

    The trainer can add hooks to this module to log the tensors.

    It's more general than logging, but so far only used for logging.

    The forward's label argument will be used as the key in the log. If there
    is no label, then the module's label will be used.
    """

    def __init__(self, label: Optional[str] = None):
        super().__init__()
        self.label = label

    def forward(self, x, label: Optional[str] = None):
        return x


class MonitoringBase(nn.Module):
    """Example model that enables monitoring."""

    def modules_to_inspect(self) -> Dict[str, nn.Module]:
        """If this method is present, and train() is run with activations or
        weight monitoring enabled, then the modules in this list will have
        hooks attached for monitoring purposed."""
        return {}


class TensorLogger(Callback):
    "TODO: add options for switching between means and histograms." ""

    tag_class = TensorTag

    def __init__(self, tb_logger, steps_til_log=50):
        self.model = None
        self.step = 0
        self._is_enabled = False
        self.tb_logger = tb_logger
        self.last_log = -math.inf
        self.steps_til_log = steps_til_log
        self._hooks = []

    def before_train(self, trainable: Trainable, step: int):
        self.model = trainable.model
        self._add_hook(self.model)

    def before_batch(self, trainable: Trainable, step: int):
        self.step = step
        self._is_enabled = self.step - self.last_log > self.steps_til_log

    def after_batch(self, trainable: Trainable, step: int):
        self.step = step
        if self._is_enabled:
            self.last_log = self.step
        self._is_enabled = False

    def before_eval(self, trainable: Trainable, step: int):
        self._is_enabled = False

    def after_eval(self, trainable: Trainable, step: int):
        self._is_enabled = True

    def _log_hist(self, value, label):
        self.tb_logger.add_histogram(label, value, self.step)

    def _log_mean_sd(self, value, label):
        self.tb_logger.add_scalar(f"{label}_mean", value.mean(), self.step)
        self.tb_logger.add_scalar(f"{label}_std", value.std(), self.step)
        if self.step == 0:
            # The first step is very special, as we are eager for the mean and
            # sd of activations to be 0 and 1.
            warning_tol = 0.2
            m, v = value.mean(), value.var()
            log_as_warn = abs(m) > warning_tol or abs(v - 1) > warning_tol
            if log_as_warn:
                _logger.warning(
                    f"Model initialization. {label}. " f"Mean: {m}, Var: {v}"
                )
            else:
                _logger.info(
                    f"Model initialization. {label}. "
                    f"Mean: {value.mean()}, Var: {value.var()}"
                )

    def _log(self, value, label):
        if self._is_enabled:
            self._log_hist(value, label)

    @torch.no_grad()
    def _on_forward(self, module, args, kwargs, output) -> None:
        """
        Return None when the output is not modified
        """
        # args[0] is the passthrough tensor, x (same as output).
        label = kwargs.get("label", args[1]) if len(args) > 1 else None
        no_label_arg = label is None or label == ""
        if no_label_arg:
            assert hasattr(module, "label"), "TensorTag must have a label."
            label = module.label
        self._log(output, label)

    def _add_hook(self, model):
        def _add_hook(submodule):
            if isinstance(submodule, self.tag_class):
                self._hooks.append(
                    submodule.register_forward_hook(
                        self._on_forward, with_kwargs=True
                    )
                )

        model.apply(_add_hook)

    def _remove_hooks(self):
        for hook in self._hooks:
            hook.remove()

    def __del__(self):
        self._remove_hooks()


class TrainingTimers:
    """Collect timers here for convenience."""

    def __init__(self):
        self.batch = _logging.Timer()
        self.epoch = _logging.Timer()
        self.validation = _logging.Timer()
        self.recovery = _logging.Timer()

    @staticmethod
    def create_and_start():
        timer = TrainingTimers()
        timer.batch.restart()
        timer.epoch.restart()
        timer.recovery.restart()
        return timer


class EarlyStopper:
    """
    Your default choice should be to not use early stopping, especially when
    using cyclic learning rates. In lr upswings, it is normal to see the loss
    increase. Having said that, when doing experiments, early stopping can help
    drastically decreate compute requirements.

    When using early stopping for investigations, the best to customize it to
    your specific situation by make educated guesses about specific loss
    patterns and how they relate to whatever your end outputs are, such as
    matplotlib plots with vmin and vmax at some known levels.

    For example, when trying to make statements about dataset sizes, the tiny
    datasets are expected to overfit very quickly, and so we can use that
    information to de-risk a very agressive early stopping.
    """

    def __init__(
        self,
        min_epochs=0,
        min_steps=0,
        epoch_patience=0,
        step_patience=0,
        eval_patience=0,
        best_factor=1.0,
        w=0.8,
    ):
        self.min_epochs = min_epochs
        self.min_steps = min_steps
        self.epoch_patience = epoch_patience
        self.step_patience = step_patience
        self.eval_patience = eval_patience
        self.best_factor = best_factor
        self.w = w
        self.reset()

    def __str__(self):
        s = f"{self.__class__.__name__}("
        s += f"min_epochs={self.min_epochs}, "
        s += f"min_steps={self.min_steps}, "
        s += f"epoch_patience={self.epoch_patience}, "
        s += f"step_patience={self.step_patience}, "
        s += f"eval_patience={self.eval_patience}, "
        s += f"best_factor={self.best_factor}, "
        s += f"w={self.w})"
        return s

    def reset(self):
        """Allow a single EarlyStopper object to be reused."""
        self.best_sloss = None
        self.best_epoch = 0
        self.best_step = 0
        self.best_eval = 0
        self.current_eval = -1
        self.smooth_loss = None
        self.continuous_infs = 0
        self.infs_threshold = 4

    def _w(self):
        # Don't let w go below 1/n_evals. This makes faster updates at the
        # beginning. current_eval is the idx, and can be 0.
        n_evals = self.current_eval + 1
        w = max(1 / n_evals, self.w)
        return w

    def update(self, loss, epoch, step):
        def _update():
            self.current_eval += 1
            if self.smooth_loss is None:
                self.smooth_loss = loss
                self.best_sloss = loss
                self.best_epoch = epoch
                self.best_step = step
            else:
                w = self._w()
                self.smooth_loss = w * self.smooth_loss + (1 - w) * loss
            if math.isinf(loss):
                self.continuous_infs += 1

        _update()

        assert self.smooth_loss is not None
        # If found a new best, reset the markers. Don't stop early.
        if self.smooth_loss < self.best_sloss:
            _logger.info(
                f"[EarlyStopper] New best loss. {self.smooth_loss:.3e} < "
                f"{self.best_sloss:.3e}. Don't stop."
            )
            self.best_sloss = self.smooth_loss
            self.best_epoch = epoch
            self.best_step = step
            self.best_eval = self.current_eval
            do_early_stop = False
        # If we have too many infs, stop early.
        elif self.continuous_infs > self.infs_threshold:
            _logger.info(
                f"[EarlyStopper] {self.infs_threshold} sequential infs. Stop."
            )
            do_early_stop = True
        # Otherwise, check if we haven't seen improvement in a while.
        elif self.smooth_loss > self.best_sloss * self.best_factor:
            is_warmup = (epoch < self.min_epochs) or (step < self.min_steps)
            # epoch patience = 0 means just 1 epoch of no improvement causes
            # early stopping. Hence the >
            n_epochs_since_best = epoch - self.best_epoch
            n_steps_since_best = step - self.best_step
            n_evals_since_best = self.current_eval - self.best_eval
            epoch_stop = n_epochs_since_best > self.epoch_patience
            step_stop = n_steps_since_best > self.step_patience
            eval_stop = n_evals_since_best > self.eval_patience
            do_early_stop = (not is_warmup) and (
                epoch_stop and step_stop and eval_stop
            )
            if do_early_stop:
                _logger.info(
                    f"[EarlyStopper] No improvement for "
                    f"{n_epochs_since_best} epochs, "
                    f"{n_steps_since_best} steps, "
                    f"{n_evals_since_best} evals. Stop."
                )
        else:
            do_early_stop = False
        return do_early_stop


def configure_cuda():
    # Enable Cuda convolution auto-tuning. For more details, see:
    #   https://pytorch.org/tutorials/recipes/recipes/tuning_guide.html
    torch.backends.cudnn.benchmark = True

    # Create a GradScalar. For more details, see:
    #   https://pytorch.org/docs/stable/notes/amp_examples.html#amp-examples
    grad_scaler = torch.amp.GradScaler("cuda")
    return grad_scaler


def create_optimizer(model_params, lr, weight_decay):
    """
    No need to abstract the optimizer creation into the interface yet; however,
    it's used in two places now (train & lr_find), so at least deduplicate it
    to avoid insidious bugs.
    """
    res = torch.optim.AdamW(
        # Why use lr with AdamW, if we use a scheduler?
        # Because the scheduler isn't used for the first epoch? So start
        # slow, lr=lr/25.
        model_params,
        lr=lr,
        # Default eps is 1e-8, which can give us gradients in excess of 1e8.
        # Setting it lower, inspired by Jeremy Howard, to say 1e-5 can
        # prevent this.
        # eps=1e-5,
        weight_decay=weight_decay,
        # Beta default is (0.9, 0.999). Here, we instead follow Jeremy Howard
        # into the dark world of hyperparameter tuning and use (0.9, 0.99).
        # See: https://www.fast.ai/posts/2018-07-02-adam-weight-decay.html
        betas=(0.9, 0.99),
    )
    return res




def inspect_scheduler(
    scheduler_fn: Callable[
        [torch.optim.Optimizer], torch.optim.lr_scheduler.LRScheduler
    ],
    n_steps: int,
) -> Tuple[float, float, float, plotly.graph_objects.Figure]:
    dummy_parameters = torch.nn.Parameter(torch.zeros(1))
    dummy_optm = torch.optim.AdamW([dummy_parameters], lr=1e-3)
    scheduler = scheduler_fn(dummy_optm)
    lrs = []
    # Some schedulers complain if optimizer.step() is not called before the
    # scheduler step.
    dummy_optm.step()
    for i in range(n_steps):
        scheduler.step()
        lr = dummy_optm.param_groups[0]["lr"]
        lr2 = scheduler.get_last_lr()[0]
        assert lr == lr2
        lrs.append(lr)
    min_lr = min(lrs)
    max_lr = max(lrs)
    last_lr = lrs[-1]
    return min_lr, max_lr, last_lr, lrs


SchedulerFn: TypeAlias = Callable[
    [torch.optim.Optimizer], torch.optim.lr_scheduler.LRScheduler
]


def train(
    trainable: Trainable,
    n_epochs: int,
    batch_size: int,
    lr: float,
    weight_decay: float,
    out_dir: Union[str, pathlib.Path],
    save_checkpoints: bool = True,
    steps_til_log: int = 1000,
    steps_til_eval: Optional[int] = None,
    evals_til_eval_train_ds: Optional[int] = None,
    early_stopper: Optional[EarlyStopper] = None,
    initial_checkpoint: Optional[Union[str, pathlib.Path]] = None,
    logger: Optional[Any] = None,
    n_workers: int = 4,
    pin_memory: bool = False,
    log_activations: bool = False,
    scheduler_fn: Optional[SchedulerFn] = None,
    callbacks: Optional[List[Callback]] = None,
    eval_batch_size: Optional[int] = None,
    max_n_checkpoints: int = 1,
    fuse_adam: bool = True,
    samples_per_epoch: Optional[int] = None,
):
    """
    Train a model.

    This is a training loop that works with any Trainable object.

    It encapsulates basic functionality like logging, checkpointing and
    choosing when to run an evalutaion. Users might be just as well off
    by copying the code to use as a baseline and modifying it to their needs.
    """
    logging.info(f"Training {trainable.label}")
    torch.set_float32_matmul_precision("high")
    out_dir = pathlib.Path(out_dir)
    # Setup output (logging & checkpoints).
    if not logger:
        tensorboard_dir = out_dir / "tensorboard"
        logger = _logging.TbLogger(tensorboard_dir)

    eval_batch_size = eval_batch_size or batch_size

    if callbacks is None:
        callbacks = []
    if log_activations:
        callbacks.append(TensorLogger(logger.writer))

    # Load the model & loss fn.
    # The order here is important when resuming from checkpoints. We must:
    # 1. Create model & log structure
    #     - logging the model summary has a secondary function of testing the
    #       model integrity: it will fail if the model cannot be constructed.
    #       We want to do this test before the time consuming step of loading
    #       the dataset.
    # 2. Send model to target device
    # 3. Load datasets
    # 4. Create optimizer & scheduler
    #     - these can be present in checkpoints.
    #     - the dataset length must be known to construct the initial scheduler.
    # 5. Initialize from checkpoint
    #     - last, after we have the model, optimizer and scheduler created.
    #
    # Further details:
    # Another reason the order is crucial is that the optimizer must be on the
    # gpu before having it's parameters populated, as there is no optimizer.gpu()
    # method (possibly coming: https://github.com/pytorch/pytorch/issues/41839).
    # An alternative would be to use the map_location argument. See the
    # discussion: https://github.com/pytorch/pytorch/issues/2830.
    model = trainable.model
    model.cuda()
    grad_scaler = configure_cuda()
    # Allow models to specify parameters with and without weight decay.
    if hasattr(model, "optim_param_groups"):
        with_decay, without_decay = model.optim_param_groups()
        params = [
            {"params": with_decay, "weight_decay": weight_decay},
            {"params": without_decay, "weight_decay": 0.0},
        ]
    else:
        params = model.parameters()
    optimizer = torch.optim.AdamW(
        # Why use lr with AdamW, if we use a scheduler?
        # Because the scheduler isn't used for the first epoch? So start
        # slow, lr=lr/25.
        params,
        lr=lr / 25,
        weight_decay=weight_decay,
        # Beta default is (0.9, 0.999). Here, we instead follow Jeremy Howard
        # into the dark world of hyperparameter tuning and use (0.9, 0.99).
        # See: https://www.fast.ai/posts/2018-07-02-adam-weight-decay.html
        betas=(0.9, 0.99),
        fused=fuse_adam,
    )

    model.train()
    # Before going any further, log model structure.
    out_file = out_dir / "model_summary.txt"
    with open(out_file, "w", encoding="utf-8") as f:
        # This is allowed to fail, as if the model has issues, we want to
        # get the errors from the actual training forward pass.
        summary = None
        try:
            summary = trainable.model_summary(batch_size=batch_size)
        except Exception as e:
            msg = (
                "Failed to generate model summary. Exception raised:\n"
                f"{str(e)}"
            )
            _logger.error(msg)
            summary = msg
        f.write(str(summary))

    metric_tracker = _logging.MetricTracker(out_dir)

    # Load the data.
    train_dl_fn, val_dl_fn = _create_dataloaders(
        batch_size=batch_size,
        eval_batch_size=eval_batch_size,
        n_workers=n_workers,
        pin_memory=pin_memory,
        persistent_workers=False,
        samples_per_epoch=samples_per_epoch,
    )
    # Go ahead and create the train_dl.
    # old: train_dl = train_dl_fn(trainable.train_ds())
    # This is part of BaseTrainable, and I guess it should be in all.
    if hasattr(trainable, "train_dl"):
        train_dl = trainable.train_dl(train_dl_fn)
    else:
        train_dl = train_dl_fn(trainable.train_ds())
    if samples_per_epoch is not None:
        assert len(train_dl) == 1 + (samples_per_epoch - 1) // batch_size
    total_n_steps = n_epochs * len(train_dl)

    # Scheduler is down here, after the dataset is loaded (to check size).
    split_percent = np.array(
        [
            len(trainable.train_ds()),
            len(trainable.val_ds()),
            len(trainable.test_ds()),
        ]
    )
    split_percent = np.round(100 * split_percent / split_percent.sum(), 1)
    # Log dataset details.
    _logger.info(
        "Dataset lengths:\n"
        f"\t{'train:':<6} {len(trainable.train_ds()):,}\n"
        f"\t{'val:':<6} {len(trainable.val_ds()):,}\n"
        f"\t{'test:':<6} {len(trainable.test_ds()):,}\n"
        f"Split ratio: {split_percent}"
    )
    # Log train loop details.
    _logger.info(
        "Train loop:\n"
        f"\tbatch size: {batch_size}\n"
        f"\teval batch size: {eval_batch_size}\n"
        f"\tepochs: {n_epochs}\n"
        f"\tsteps per epoch: {len(train_dl):,}\n"
        f"\ttotal steps: {total_n_steps:,}"
    )

    def _scheduler_fn(_optimizer):
        return torch.optim.lr_scheduler.OneCycleLR(
            _optimizer,
            max_lr=lr,
            # Either specify steps & epochs, or specify total steps manually.
            # steps_per_epoch=len(train_dl),
            # epochs=n_epochs,
            total_steps=total_n_steps,
            # Testing
            three_phase=True,
        )

    scheduler_fn = _scheduler_fn if scheduler_fn is None else scheduler_fn
    scheduler = scheduler_fn(optimizer)
    lr_min, lr_max, lr_last, lrs = inspect_scheduler(
        scheduler_fn, total_n_steps
    )
    _logging.log_scheduler(
                lr_min, lr_max, lr_last, scheduler.__class__.__name__, lrs,
        "lr_schedule.html"
    )
    # Record how many times schedule step is skipped due to grad scalar backoff.
    # We want to track this as too many steps will cause the resulting lr
    # schedule clipping to be significant.
    n_lr_steps_skipped = 0
    step = 0
    # Defined here as it is shared between both epoch and batch loops.
    stop_early = False

    model_saver = _logging.ModelSaver(
        out_dir,
        trainable.model,
        optimizer,
        scheduler,
        max_history=max_n_checkpoints,
    )

    if initial_checkpoint is not None:
        _logging.load_model_and_optimizer(
            model, initial_checkpoint, optimizer, scheduler
        )

    def _eval(use_train_ds: bool = False):
        for cb in callbacks:
            cb.before_eval(trainable, step)
        label = "train-ds" if use_train_ds else "val-ds"
        _logger.info(f"Running evaluation {label}")
        with evaluating(model), torch.no_grad(), timers.validation:
            if use_train_ds:
                eval_results = trainable.evaluate_train(train_dl_fn)
            else:
                eval_results = trainable.evaluate_val(val_dl_fn)
        logger.log(step, eval_results, label)
        if "metrics" not in eval_results:
            raise ValueError("Trainable.evaluate() must return metrics.")
        metrics = eval_results["metrics"]
        assert (
            metrics[0].name == "loss"
        ), "Currently, by convention, the first metric must be loss."
        _logging.print_metrics(metrics)
        logger.log_scalar(step, "eval-time", timers.validation.elapsed(), label)
        _logger.info(
            f"Finished evaluation in {round(timers.validation.elapsed())} sec "
            f"(rolling ave: {round(timers.validation.rolling_duration())} sec)"
        )
        for cb in callbacks:
            cb.after_eval(trainable, step)
        return metrics

    n_evals = 0

    def _eval_and_checkpoint():
        stop_early = False
        metrics = _eval()
        nonlocal n_evals  # yes, another sign that a class would give benefits.
        n_evals += 1
        if evals_til_eval_train_ds and n_evals % evals_til_eval_train_ds == 0:
            _eval(use_train_ds=True)

        assert (
            metrics[0].name == "loss"
        ), "By convention, the first metric must be loss."
        # If this on_metric_end type of behaviour grows, consider switching
        # to callbacks.
        # Using step (not batch_step).
        improved_metrics = metric_tracker.new_metrics(
            metrics,
            epoch,
            step,  # batch_step
        )
        if save_checkpoints:
            model_saver.save_step_checkpoint(step, improved_metrics)
            timers.recovery.restart()
        if early_stopper:
            should_stop = early_stopper.update(metrics[0].value, epoch, step)
            if should_stop:
                _logger.info(f"Early stopping triggered. {str(early_stopper)}")
                stop_early = True
        return stop_early

    _logger.info("Starting training loop.")
    timers = TrainingTimers.create_and_start()
    for cb in callbacks:
        cb.before_train(trainable, step)
    # Do an initial eval.
    _eval()
    # The length of the epoch inner loop is now clearly too long to be able
    # to understand it at a glance.
    # Meters to calculate smooth values for terminal logging.
    beta = max(1e-3, 1 / steps_til_log)
    loss_meter = _logging.MovingAverageMeter(beta, name="loss")
    lr_meter = _logging.MovingAverageMeter(beta, name="lr")
    model_mean, model_sd, grad_norm = (0, 0, 0)  # predefine for logging.
    for epoch in range(n_epochs):
        _logger.info(f"Starting epoch {epoch+1}/{n_epochs}\t({out_dir})")
        timers.epoch.restart()
        # fmt: off
        _logging.log_step(epoch, n_epochs, -1, len(train_dl), timers.epoch, timers.batch, loss_meter.avg, lr_meter.avg, model_mean, model_sd, grad_norm)
        # fmt: on
        for batch_step, sample in enumerate(train_dl):
            if batch_step > len(train_dl):
                import pdb

                pdb.set_trace()
                raise ValueError("Batch step exceeds length of train_dl.")
            for cb in callbacks:
                cb.before_batch(trainable, step)
            timers.batch.restart()
            # set_to_none=True is suggested to improve performance, according to:
            #   https://pytorch.org/tutorials/recipes/recipes/tuning_guide.html
            # A recipe for autocast, grad scaling and set_to_none:
            #   https://pytorch.org/tutorials/recipes/recipes/amp_recipe.html
            optimizer.zero_grad(set_to_none=True)
            with torch.autocast("cuda"):
                model_out, total_loss = trainable.forward(sample)
            grad_scaler.scale(total_loss).backward()
            ## Optional clipping.
            grad_scaler.unscale_(optimizer)
            MAX_NORM = 10000  # pretty much disabled. Kept around for logging.
            grad_norm = torch.nn.utils.clip_grad_norm_(
                model.parameters(), MAX_NORM
            )
            # Don't bother warning if infs are found, as the grad scaler
            # will skip the optimizer step anyway.
            grad_has_infs = grad_norm == float("inf")
            if not grad_has_infs:
                if grad_norm > MAX_NORM:
                    _logger.warning(
                        f"Gradients clipped. Norm: {grad_norm:.1e} > {MAX_NORM}"
                    )
            ## \Optional
            # dbg = {pn:(p, p.stride(), p.grad.stride() if p.grad is not None else 0) for pn, p in model.named_parameters() if p.grad is not None and p.grad.stride() != p.stride()}
            grad_scaler.step(optimizer)
            # Recording the before and after scale. Why? The grad scaler can
            # create NaN gradients in the first few iterations, and when it
            # does, it skips the updating of the optimizer. We want to
            # know when this happens so that we can also skip the learning
            # rate scheduler update.
            # ref: https://discuss.pytorch.org/t/optimizer-step-before-lr-scheduler-step-error-using-gradscaler/92930/10
            scale_before = grad_scaler.get_scale()
            grad_scaler.update()
            assert grad_scaler.get_backoff_factor() < 1.0, (
                "The logic for skipping the learning rate scheduler "
                "relies on the backoff factor being less than 1.0, which "
                "it is the default (0.5)."
            )
            scale_has_decreased = scale_before > grad_scaler.get_scale()
            skip_lr_sched = scale_has_decreased
            # get_last_lr() returns a list, but we only have one param group.
            last_lr = scheduler.get_last_lr()[0]
            lr_meter.update(last_lr)

            metrics = [
                _logging.Metric("loss", total_loss.item()),
            ]
            logger.log_metrics(step, metrics, log_group="train")
            logger.log_scalar(step, "epoch", epoch, log_group="train")
            logger.log_scalar(step, "lr", last_lr, log_group="train")
            logger.log_scalar(step, "grad_norm", grad_norm, log_group="train")

            # We log total_loss directly to logging framework, but log a
            # smoothed loss to the console.
            loss_meter.update(total_loss.item())
            # Log step. Periodically, and at the end of the epoch.
            is_last = batch_step + 1 == len(train_dl)
            # step + 1, as steps are displaed 1-indexed (so ahead).
            if (step > 0 and (step + 1) % steps_til_log == 0) or is_last:
                model_mean = torch.mean(model_out).item()
                model_sd = torch.std(model_out).item()
                _logging.log_step(
                    epoch,
                    n_epochs,
                    batch_step,
                    len(train_dl),
                    timers.epoch,
                    timers.batch,
                    loss_meter.avg,
                    lr_meter.avg,
                    model_mean,
                    model_sd,
                    grad_norm,
                )

            # Evaluate.
            # batch_step + 1, as steps are displayed 1-indexed (so ahead).
            # If switching back to (batch_step), make sure to skip 0.
            if steps_til_eval and (batch_step + 1) % steps_til_eval == 0:
                is_near_epoch_end = batch_step + steps_til_eval >= len(train_dl)
                if not is_near_epoch_end:
                    stop_early = _eval_and_checkpoint()
                    if stop_early:
                        break
            step_matches = (
                step == epoch * len(train_dl) + batch_step - n_lr_steps_skipped
            )
            if not step_matches:
                import pdb

                pdb.set_trace()

            # assert (
            #     step == epoch * len(train_dl) + batch_step - n_lr_steps_skipped
            # ), f"step: {step}, epoch: {epoch}, batch_step: {batch_step}"
            if not skip_lr_sched:
                if grad_has_infs:
                    raise ModelException(
                        "If infs found, step should be skipped. Might be in a "
                        "GradScaler death spiral (probably caused by your model "
                        "producing infs/NaNs)."
                    )
                if not torch.isfinite(total_loss):
                    raise ModelException(
                        "Loss is not finite. This is likely a model error."
                    )
                scheduler.step()
                # Option: do or [don't] increment step if grad scaler skipped.
                step += 1
            else:
                n_lr_steps_skipped += 1
                _logger.info(
                    f"GradScaler didn't step (step {step}). "
                    f"Total skipped: {n_lr_steps_skipped} "
                    f"({round(float(n_lr_steps_skipped)/(step+1)):>3.0%}). "
                    f"New scale: {grad_scaler.get_scale():.3f}"
                )
            # Recovery.
            # Don't allow training to proceed too long without checkpointing.
            if timers.recovery.elapsed() > RECOVERY_CKPT_PERIOD_SEC:
                model_saver.save_recovery()
                timers.recovery.restart()
            for cb in callbacks:
                cb.after_batch(trainable, step)
            # One more batch to dust!

        _logger.info(
            f"Finished epoch in {round(timers.epoch.elapsed())} secs "
            f"(rolling ave (train + eval): "
            f"{round(timers.epoch.rolling_duration())} s/epoch)"
        )
        # Evaluate and save at end of epoch. We may have already evaluated 
        # and decided to stop early, so check that first.
        stop_early = stop_early or _eval_and_checkpoint()
        if stop_early:
            break
    for cb in callbacks:
        cb.after_train(trainable, step)
    _logger.info(
        f"Finished training. {round(timers.epoch.total_elapsed())} "
        "secs elsapsed."
    )
    _logger.info(f"out_dir: {out_dir}")
    # If persistent workers is enabled, then when a dataloader is consumed,
    # it does not clean up it's workers, as they may be reused. I think this
    # leaves th deletion trigger to be delayed until Python's garbage
    # collection runs. This may happen some time later. If you are running
    # train() in a loop, then the deletion of workers might be delayed until
    # after the next dataloader is created. This could also happen if you hang
    # on to a reference to the dataloader after training. The BaseTrainable
    # caches dataloaders, so this could indeed be a problem!
    # A fix (https://github.com/pytorch/pytorch/pull/39869) that was made to
    # Pytorch to deal with hangning workers (why they might hang seems to be a
    # mystery), which checks all workers to see if they are alive. The source
    # for is_alive() is at:
    # https://github.com/python/cpython/blob/6fc643674983e27ec5cc312f2e83468050d1d364/Lib/multiprocessing/process.py#L153
    # The assert that we sometimes see fails because the process calling
    # worker.is_alive() is not the process that created the worker.



import kddl
import kddl.train as train
import polars as pl
import math
import logging
from typing import Tuple, Optional, Sequence, List
import numpy as np
from numpy.lib.stride_tricks import sliding_window_view
import scipy.stats
import itertools

_logger = logging.getLogger(__name__)


def lr_sched(lr_min, lr_max, n_lr_steps):
    sched = lr_min * (lr_max / lr_min) ** np.linspace(0, 1, n_lr_steps)
    return sched


def lr_early_stopper(n_steps):
    return train.EarlyStopper(
        min_steps=n_steps // 2,
        step_patience=5,
        best_factor=10,
    )


class TroublesomeLrCurve(Exception):
    """The LR curve is problematic. This is a case worth investigating."""

    def __init__(self, message):
        super().__init__(message)
        self.message = message


def sweep(
    trainable,
    lr_min,
    lr_max,
    n_lr_steps,
    n_runs,
    early_stopper=None,
    init_weights=None,
    log_space=True,
    **train_kwargs,
) -> Tuple[List[List[float]], List[float]]:
    """Run an LR sweep `n_runs` times, each time starting from `init_weights`.

    As the sweep can terminate early, the losses are returned as a list of
    lists, potentially of different lengths.
    """
    _logger.info(
        f"{n_runs} LR sweeps from {lr_min:.1e} to {lr_max:.1e} in "
        f"{n_lr_steps} steps."
    )
    _logger.info(f"Train kwargs: {train_kwargs}")
    if init_weights is None:
        init_weights = trainable.model.state_dict()
    if log_space:
        lr_sched = (
            lr_min * (lr_max / lr_min) ** np.linspace(0, 1, n_lr_steps)
        ).tolist()
    else:
        lr_sched = np.linspace(lr_min, lr_max, n_lr_steps).tolist()
    all_losses = []
    for run in range(n_runs):
        if early_stopper is not None:
            # Don't reuse the same early stopper each run!
            early_stopper.reset()
        trainable.model.load_state_dict(init_weights)
        _, losses = train._lr_sweep(
            trainable, lr_sched, early_stopper=early_stopper, **train_kwargs
        )
        all_losses.append(losses)
        best_idx = np.argmin(losses)
        best_lr = lr_sched[best_idx]
        best_loss = losses[best_idx]
        _logger.info(
            f"Run {run} [finished]\tBest (step, lr, loss) "
            f"= ({best_idx}, {best_lr:.1e}, {best_loss:.3g})"
        )
    return all_losses, lr_sched


def steepest_loss(losses, run_lens, lr_sched, k_smooth):
    if k_smooth % 2 == 0:
        raise ValueError("k_smooth must be odd.")
    mask = (
        np.stack([np.arange(len(lr_sched)) for i in range(len(run_lens))])
        < run_lens[:, None]
    )
    losses = losses * mask
    # Shift right by 1 then re-mask, as otherwise we diff to first mask zero.
    # We want 0 to be the first value too.
    dloss = np.where(mask, np.pad(np.diff(losses), ((0, 0), (1, 0))), 0)
    mask_sum = np.sum(mask, axis=0)
    ave_dloss = np.divide(np.sum(dloss, axis=0), mask_sum, where=mask_sum > 0)
    # TODO: the lrs are evenly spaced in log-space, so the equal weighting
    # by smoothing could be improved. The end goal would be to use Kalman
    # filtering+smoothing to estimate the dloss. For now, smoothing.
    smooth_dloss = np.convolve(
        ave_dloss, np.ones(k_smooth) / k_smooth, mode="valid"
    )
    smooth_dloss = np.pad(smooth_dloss, (k_smooth // 2, k_smooth // 2))
    # TODO: need a better way to cut of the possibly noisy end.
    # Don't go past the max run length, or 75% of the full length.
    max_idx = min(np.min(run_lens), math.floor(len(lr_sched) * 0.75))
    steepest_idx = np.argmin(smooth_dloss[0:max_idx])
    steepest_lr = lr_sched[steepest_idx]
    if np.any(steepest_idx >= run_lens):
        _logger.error(
            f"Steepest loss at {steepest_idx} is beyond the run length {run_lens}"
        )
        # Fix is to better cut the noisy end.
    # assert all(steepest_idx < run_lens)
    return steepest_lr, steepest_idx, ave_dloss, smooth_dloss


def pad_over_inf(w: List[float], to_len: int) -> np.ndarray:
    w = np.array(w)
    nonfinite = np.flatnonzero(~np.isfinite(w))
    if len(nonfinite) == 0:
        return w
    first_invalid = nonfinite[0]
    if first_invalid == 0:
        return np.full(to_len, fill_value=0)
    w = np.array(w)[:first_invalid]
    max_val = np.max(w)
    assert np.isfinite(max_val)
    return np.pad(
        w,
        (0, to_len - first_invalid),
        mode="constant",
        constant_values=max_val,
    )


def is_flat_entropy_test(loss_arr, min_tol=1.3):
    """
    Args:
        loss_arr: 2D array of losses, with shape (n_runs, n_steps)
        min_tol: entropy_all - entropy_init < np.log(min_tol) => flat
            The lower the value, the harder it is to be considered flat.
    """
    if min_tol <= 1:
        raise ValueError("min_tol must be > 1.")
    INIT_LEN = 50
    at_start = loss_arr[:, :INIT_LEN].flatten()
    mu = np.mean(at_start)
    # A recent change, doubling the variance. Would have been good to have
    # more test cases to ensure this heuristic is still okay.
    var = np.var(at_start) #* 2
    entropy_init = -np.mean(scipy.stats.norm.logpdf(at_start, mu, var))
    entropy_all = -np.mean(scipy.stats.norm.logpdf(loss_arr, mu, var))
    is_flat = entropy_all - entropy_init < np.log(min_tol)
    return is_flat


def is_flat(loss_arr, min_tol=1.3):
    by_entropy = is_flat_entropy_test(loss_arr, min_tol)
    # Any other tests?
    return by_entropy


def clip_to_finite(w: Sequence[float]):
    for i, v in enumerate(w):
        if not np.isfinite(v):
            return w[:i]
    return w


def to_loss_arr(losses: Sequence[Sequence[float]]) -> Optional[np.ndarray]:
    loss_arr = [clip_to_finite(loss) for loss in losses]
    max_len = min(len(l) for l in loss_arr)
    loss_arr = np.stack([np.array(l)[0:max_len] for l in loss_arr])
    assert np.all(np.isfinite(loss_arr)), "Should be no infs/nans"
    if loss_arr.size == 0:
        return None
    # loss_arr[np.isnan(loss_arr)] = np.inf
    return loss_arr


def cut_takeoff(loss_arr):
    """Cut the right-side if the losses are very large.

    The losses can grow massively towards the end of the lr-find. This
    interferes with the kalman smoothing. We cut off the right-side if the
    losses are above the starting loss.

    Two options here:
      1. Take until the last point that is below the limit.
         An issue with this option is that the loss can spike and then come
         back down again, and this criteria can cause the spike to be
         included.
      2. Take until the first point that is above the limit.
         This is more agressive, but avoids the above issue. To balance the
         aggressiveness, we can increase the quantile used in the above_limit.
    """
    n_runs, n_steps = loss_arr.shape
    assert n_runs < n_steps, "The axis are wrong."

    def is_too_high(quantile):
        """Choose a good representative low value, and get it's index."""
        # The amount of smoothing will effect the choice of quantile.
        k = 15
        long_enough_to_ave = n_steps > k * 5
        if long_enough_to_ave:
            smoothed_loss = np.convolve(
                loss_arr.mean(axis=0), np.ones(k) / k, mode="valid"
            )
            smoothed_loss = np.pad(smoothed_loss, (k // 2, k // 2), mode="edge")
            assert len(smoothed_loss) == loss_arr.shape[1]
            low_val = smoothed_loss.min()
            low_at = np.flatnonzero(smoothed_loss <= low_val)
        else:
            _logger.warning("loss array too short to smooth.")
            smoothed_loss = loss_arr.mean(axis=0)
            low_val = smoothed_loss.min()
            low_at = np.flatnonzero(smoothed_loss <= low_val)
        low_at = n_steps if len(low_at) == 0 else low_at[0]
        assert 0 <= low_at < n_steps
        # Attempts to give a little padding seem to generally fail. There can
        # be sudden jumps, which negatively affect Kalman smoothing results.
        if not low_at > 0:
            _logger.error("Expected +ve. is_flat() should have caught this.")
            raise TroublesomeLrCurve(
                "Curve doesn't descend, yet isn't detected as flat."
            )
        # ignore_over = np.quantile(smoothed_loss[0:low_at], quantile)
        # Switch to band, as long initial slopes can allow large swings to be
        # included.
        band_len = 100
        band_around_min = loss_arr[:, max(0, low_at - band_len) : low_at]
        ignore_over = np.quantile(band_around_min, quantile)
        _logger.debug(f"{ignore_over=:.3f}")
        # ignore_over = low_val
        above_limit = smoothed_loss > ignore_over
        return low_at, low_val, above_limit

    if is_flat(loss_arr):
        return loss_arr

    low_at, low_val, above_limit = is_too_high(quantile=0.95)
    is_after_min = np.zeros(n_steps, dtype=bool)
    # Don't cut off before min is reached.
    # Attempts to give a little padding seem to generally fail. The min can be
    # very late, and very close to sudden jumps, which negatively affect Kalman
    # smoothing results, so be careful doing this.
    buff = 0
    is_after_min[low_at + buff + 1 :] = True
    above_max = np.flatnonzero(is_after_min & above_limit)
    max_t = loss_arr.shape[1] if len(above_max) == 0 else above_max[0]
    _logger.debug(f"Cutting off at {max_t} steps. ({low_val=:.3f})")
    assert np.all(
        np.isfinite(loss_arr[:, 0 : max_t + 1])
    ), "Should be no infs/nans"
    loss_arr = loss_arr[:, 0 : max_t + 1]
    return loss_arr


def mean_smooth(losses: Sequence[Sequence[float]], k: int = 21):
    losses = cut_takeoff(losses)
    dloss = np.pad(np.diff(losses), ((0, 0), (1, 0)))
    mean_dloss = dloss.mean(axis=0)
    k = 21
    smoothed_dloss = np.convolve(mean_dloss, np.ones(k) / k, mode="valid")
    # Pad with the edge values. Don't convolve as "same", as we want to
    # reduce flick towards 0.
    smoothed_dloss = np.pad(smoothed_dloss, (k // 2, k // 2), mode="edge")
    return mean_dloss, smoothed_dloss


def kalman_smooth(loss_arr, Q_scale: float = 1):
    # Q_scale: float = math.sqrt(2)):
    # Q_scale = math.sqrt(2)
    Q_scale = 1 / 3
    _logger.debug("[start] Kalman smoothing")
    losses = cut_takeoff(loss_arr)
    dloss = np.pad(np.diff(losses), ((0, 0), (1, 0)))

    # First, we will estimate good values for the process and observation noise
    # variance. The transition between min and max of the smoothed dloss should
    # be obtained gradually by the changing state, so (max-min) acts as an
    # upper bound on the process noise variance, Q.
    # k = 15
    # long_enough_to_ave = dloss.shape[1] > k * 5
    # if long_enough_to_ave:
    #     ave_smoothed_loss = np.convolve(
    #         losses.mean(axis=0), np.ones(k) / k, mode="valid"
    #     )
    #     est_Q = np.var(np.diff(ave_smoothed_loss)) * Q_scale
    #     # est_Q = np.var(ave_smoothed_dloss)
    #     if not np.isfinite(dloss.mean()):
    #         import pdb
    #
    #         pdb.set_trace()
    #     _logger.info(f"{est_Q=:.3e}")
    #     _logger.info(f"{dloss.mean()=:.3e}")
    # else:
    #     est_Q = 1e-4

    obs = dloss.T
    n_steps, n_obs = obs.shape
    # R_k is estimated by a sliding window variance of the observations.
    # This allows it to change over time, which is what we expect.
    window_len = min(31, ((n_steps + 1) // 2) * 2 - 1)
    vars = np.pad(
        sliding_window_view(obs, (window_len, n_obs)).var(axis=(1, 2, 3)),
        (window_len // 2, window_len // 2),
        mode="edge",
    )
    assert vars.shape == (n_steps,), f"{vars.shape=}, {n_steps=}"
    R_to_Q = 8000
    R_k = [R_to_Q / (1 + R_to_Q) * r * np.eye(n_obs) for r in vars]
    assert R_k[0].shape == (n_obs, n_obs)
    Q = [max(1e-7, 1 / (1 + R_to_Q)) * r for r in vars]
    _logger.debug(f"Q: {np.mean(Q):.3e}, {np.min(Q):.3e}, {np.max(Q):.3e}")
    # Q = [est_Q for r in vars]

    # Kalman filter parameters
    F = 1  # State transition matrix (scalar)
    H = np.ones((n_obs, 1))  # Observation matrix

    # Initial estimates
    x_hat = np.zeros(n_steps)
    # x_hat[0] = obs[0:20].mean()  # Initial estimate of delta-loss
    # x_hat = np.mean(obs, axis=1)  # Filtered estimates of delta-loss
    # x_hat = np.full(
    #     n_steps, fill_value=obs[0].mean()
    # )  # Filtered estimates of delta-loss
    P = np.zeros(n_steps)  # Filtered estimate covariance
    P[0] = 10 * Q[0]  # Initial estimate covariance
    x_hat_minus = np.zeros(n_steps)  # Predicted state estimates
    P_minus = np.zeros(n_steps)  # Predicted estimate covariance

    # Kalman filter forward pass
    for k in range(1, n_steps):
        # Prediction step
        x_hat_minus[k] = F * x_hat[k - 1]
        P_minus[k] = F * P[k - 1] * F + Q[k]

        # Update step
        z_k = obs[k].reshape(-1, 1)  # Observation vector at time k
        y_k = z_k - H * x_hat_minus[k]  # Innovation vector

        S_k = H * P_minus[k] * H.T + R_k[k]  # Innovation covariance matrix
        # S_k = S_k + 1e-8 * np.eye(S_k.shape[0])  # Add small noise to avoid singularity
        K_k = P_minus[k] * H.T @ np.linalg.inv(S_k)  # Kalman gain vector

        x_hat[k] = x_hat_minus[k] + (K_k @ y_k).item()  # Updated state estimate
        P[k] = (
            P_minus[k] - (K_k @ S_k @ K_k.T).item()
        )  # Updated estimate covariance

    # Kalman smoother backward pass
    x_hat_smooth = np.zeros(n_steps)  # Smoothed state estimates

    # Initialize with the last filtered estimates
    x_hat_smooth[-1] = x_hat[-1]
    for k in range(n_steps - 2, -1, -1):
        # Compute the smoothing gain
        # Warn if expecting numerical issues.
        A_k = P[k] * F / P_minus[k + 1]
        if abs(A_k) > 1e3:
            _logger.warning(f"{A_k=:.3e} at {k}. Possible numerical issues.")

        # Smoothed state estimate
        DAMPING = 0.95
        x_hat_smooth[k] = x_hat[k] + DAMPING * A_k * (
            x_hat_smooth[k + 1]
            - x_hat_minus[k + 1]
            # x_hat_smooth[k + 1] - x_hat[k + 1]
        )
    assert x_hat_smooth[-1] == x_hat[-1]
    return x_hat_smooth


def mean_lr_choose(lrs, loss_arr):
    mean_dloss, smoothed_dloss = mean_smooth(loss_arr)
    suggested_lr_idx_mloss = np.argmin(smoothed_dloss)
    assert suggested_lr_idx_mloss < len(lrs), suggested_lr_idx_mloss
    suggested_lr_mloss = lrs[suggested_lr_idx_mloss]
    return (
        mean_dloss,
        smoothed_dloss,
        suggested_lr_idx_mloss,
        suggested_lr_mloss,
    )


def critical_points(kalman_dloss) -> Tuple[int, int, int]:
    # Some preliminary arrays and points used in the below calculations.
    ddloss = np.concatenate([[0], np.diff(kalman_dloss)])
    cum_dloss = np.cumsum(kalman_dloss)
    min_idx = np.argmin(cum_dloss)
    if min_idx == 0:
        raise TroublesomeLrCurve("Minimum is at the start.")
        # return 0, 0, 0

    # 1. Steepest slope, not after the minimum.
    # We limit to the lhs of the minimum so as to avoid choosing points that
    # follow a regression in the loss back up.
    min_grad_idx = np.argmin(kalman_dloss[:min_idx])

    # 2. Maximum double-derivative. Acts as right bound where there can be a
    # flick up. Only consider after the min_grad_idx.
    ddloss = np.concatenate([[0], np.diff(kalman_dloss)])

    # Max ggrad
    # Order by acceleration (highest first), and take the first point that isn't
    # thrown out by the the creteria:
    #   a) it must be after the min_grad_idx
    #   b) it must be after the min_idx
    #   c) it must have a positive gradient/velocity.
    max_ggrad_idx = next(
        itertools.chain(
            itertools.dropwhile(
                lambda x: x[0] <= min_grad_idx
                or x[0] <= min_idx
                or kalman_dloss[x[0]] <= 0,
                sorted(enumerate(ddloss), key=lambda x: x[1], reverse=True),
            ),
            [(len(kalman_dloss) - 1, None)],
        )  # if no point matches above criteria.
    )[0]

    # 3. Minimum double-derivative. Acts as left bound. Only consider before
    # the min_grad_idx.
    min_ggrad_idx = np.argmin(ddloss[:min_grad_idx])

    return min_ggrad_idx, min_grad_idx, max_ggrad_idx


def kalman_lr_choose_old(lrs, loss_arr, end_margin=50):
    kalman_dloss = kalman_smooth(loss_arr)
    # suggested_lr_idx_kalman = np.argmin(kalman_dloss)
    # If end_margin is too large, clip to 1/3 of the length.
    end_margin = min(end_margin, len(kalman_dloss) // 3)
    suggested_lr_idx = next(
        itertools.dropwhile(
            lambda x: x[0] >= len(kalman_dloss) - end_margin,
            sorted(enumerate(kalman_dloss), key=lambda x: x[1]),
        )
    )[0]
    suggested_lr = lrs[suggested_lr_idx]
    assert suggested_lr_idx < len(kalman_dloss) - end_margin, suggested_lr_idx
    return kalman_dloss, suggested_lr_idx, suggested_lr


def kalman_lr_choose(lrs, loss_arr, end_margin=50):
    kalman_dloss = kalman_smooth(loss_arr)
    _, max_grad_idx, rhs = critical_points(kalman_dloss)
    # suggested_lr_idx_kalman = np.argmin(kalman_dloss)
    # If end_margin is too large, clip to 1/3 of the length.
    end_margin = min(end_margin, len(kalman_dloss) // 3)
    q = 0.5
    # Geometric weighting
    suggested_lr_idx = int(max_grad_idx**q * rhs ** (1 - q))
    # suggested_lr_idx = int((q - 1) * max_grad_idx + q * rhs)
    suggested_lr_idx = min(suggested_lr_idx, len(kalman_dloss) - end_margin)
    suggested_lr = lrs[suggested_lr_idx]
    return kalman_dloss, suggested_lr_idx, suggested_lr



import logging
import torch
import kddl.train
import kddl._logging
import torch.nn as nn
import torch.nn.functional as F
from tqdm import tqdm
import torchvision as tv

_logger = logging.getLogger(__name__)


def fashion_mnist_ds(tmp_dir=None):
    img_transform = tv.transforms.Compose(
        [
            tv.transforms.ToTensor(),
        ]
    )
    train_ds = tv.datasets.FashionMNIST(
        root=tmp_dir, train=True, download=True, transform=img_transform
    )
    test_ds = tv.datasets.FashionMNIST(
        root=tmp_dir, train=False, download=True, transform=img_transform
    )
    assert len(train_ds) == 60000 and len(test_ds) == 10000

    # Split off a validation set
    train_size, val_size = 50000, 10000
    train_ds, val_ds = torch.utils.data.random_split(
        train_ds,
        [train_size, val_size],
        generator=torch.Generator().manual_seed(123),
    )
    assert len(train_ds) == train_size and len(val_ds) == val_size

    # Check a few images
    for i in range(10):
        img, label = train_ds[i]
        assert img.shape == (1, 28, 28)
        assert 0 <= label < 10
        assert torch.all(img >= 0) and torch.all(img <= 255)
    return train_ds, val_ds, test_ds


def fashion_mnist_ds_mgr(fashion_mnist_ds):
    train_ds, val_ds, test_ds = fashion_mnist_ds
    # Extract underlying data for normalization statistics
    orig_train_ds = train_ds.dataset
    train_ds_indices = train_ds.indices
    train_imgs = orig_train_ds.data[train_ds_indices].float()
    assert train_imgs.shape == (50000, 28, 28)
    mean = torch.mean(train_imgs)
    sd = torch.std(train_imgs)
    min = torch.min(train_imgs)
    max = torch.max(train_imgs)
    res = kddl.train.BasicDatasetManager(
        train_ds,
        val_ds,
        test_ds,
        train_ds_attrs={"mean": mean, "sd": sd, "min": min, "max": max},
    )
    return res


@torch.no_grad()
def mean_0_var_1(
    forward_fn,
    input_fn,
    n_samples=int(1e2),
    mean_tol=0.1,
    var_tol=0.1,
    generator=None,
):
    """Send random data in and check the output has near mean 0 and variance 1."""
    if generator is None:
        generator = torch.Generator().manual_seed(123)
    m_out = []
    for _ in tqdm(range(n_samples)):
        inputs = input_fn()
        inputs = [i.normal_(generator=generator).cuda() for i in inputs]
        y = forward_fn(*inputs)
        m_out.append(y)
    out_shape = m_out[0].shape
    m_out = torch.stack(m_out)
    passed = True
    res = {}

    def within_tol(m, v):
        return torch.allclose(
            m, torch.zeros_like(m), atol=mean_tol
        ) and torch.allclose(v, torch.ones_like(v), atol=var_tol)

    mean = torch.mean(m_out)
    var = torch.var(m_out)
    if not within_tol(mean, var):
        passed = False

    details = (
        f"Requirement: {-mean_tol} < mean < {mean_tol}, 1 - {var_tol} < var < 1 + {var_tol}\n"
        f"Actual: mean={mean:.3e}, var={var:.3e}"
    )

    return passed, details


class Cnn1(nn.Module):
    """
    Copied from PyTorch's MNIST example: https://github.com/pytorch/examples/blob/1bef748fab064e2fc3beddcbda60fd51cb9612d2/mnist/main.py

    We did tweak the normalization.
    """

    def __init__(self):
        super().__init__()
        self.conv1 = nn.Conv2d(1, 32, 3, 1)
        self.conv2 = nn.Conv2d(32, 64, 3, 1)
        self.dropout1 = nn.Dropout(0.25)
        self.dropout2 = nn.Dropout(0.5)
        self.fc1 = nn.Linear(9216, 128)
        self.fc2 = nn.Linear(128, 10)
        self.register_buffer("mean", torch.tensor(0.0))
        self.register_buffer("sd", torch.tensor(1.0))

    def set_mean_sd(self, mean, sd):
        self.mean = mean
        self.sd = sd

    def input_norm(self, x):
        return (x - self.mean) / self.sd

    def forward(self, x):
        x = self.input_norm(x)
        x = self.conv1(x)
        x = F.relu(x)
        x = self.conv2(x)
        x = F.relu(x)
        x = F.max_pool2d(x, 2)
        x = self.dropout1(x)
        x = torch.flatten(x, 1)
        x = self.fc1(x)
        x = F.relu(x)
        x = self.dropout2(x)
        x = self.fc2(x)
        output = F.log_softmax(x, dim=1)
        return output


class FastaiCnn1(nn.Module):
    """
    Copied from Fastai's MNIST example: https://github.com/fastai/course22p2/blob/master/nbs/10_activations.ipynb
    """

    def __init__(self):
        super().__init__()
        self.register_buffer("mean", torch.tensor(0.0))
        self.register_buffer("sd", torch.tensor(1.0))
        self.layers = nn.Sequential(*self.cnn_layers())

    @staticmethod
    def conv(ni, nf, ks=3, act=True):
        res = nn.Conv2d(ni, nf, kernel_size=ks, stride=2, padding=ks // 2)
        if act:
            res = nn.Sequential(res, nn.ReLU())
        return res

    @classmethod
    def cnn_layers(cls):
        res = [
            cls.conv(1, 8, ks=5),  # 14x14
            cls.conv(8, 16),  # 7x7
            cls.conv(16, 32),  # 4x4
            cls.conv(32, 64),  # 2x2
            cls.conv(64, 10, act=False),  # 1x1
            nn.Flatten(),
        ]
        return res

    def set_mean_sd(self, mean, sd):
        self.mean = mean
        self.sd = sd

    def input_norm(self, x):
        return (x - self.mean) / self.sd

    def forward(self, x):
        x = self.input_norm(x)
        for layer in self.layers:
            x = layer(x)
        x = F.log_softmax(x, dim=1)
        return x


class ClassifyTrainable(kddl.train.Trainable):
    def __init__(self, ds_mgr, model, label):
        super().__init__(model, label)
        self.ds_mgr = ds_mgr
        self.model.set_mean_sd(ds_mgr.mean, ds_mgr.sd)

    def loss_fn(self, m_out, y):
        loss = F.nll_loss(m_out, y, reduction="mean")
        return loss

    def train_ds(self):
        return self.ds_mgr.train_ds()

    def val_ds(self):
        return self.ds_mgr.val_ds()

    def test_ds(self):
        return self.ds_mgr.test_ds()

    def forward(self, sample):
        x, y = sample
        x = x.cuda()
        y = y.cuda()
        m_out = self.model(x)
        loss = self.loss_fn(m_out, y)
        return m_out, loss

    @torch.no_grad()
    def evaluate(self, dl):
        if self.model.training:
            _logger.warning("Model should be in eval mode")
        loss_meter = kddl._logging.Meter("loss")
        acc_meter = kddl._logging.Meter("accuracy")
        for sample in dl:
            x, y = sample
            y = y.cuda()
            N = x.shape[0]
            output, loss = self.forward(sample)
            pred = output.argmax(dim=1, keepdim=True)
            n_correct = pred.eq(y.view_as(pred)).sum().item()
            acc_meter.update(n_correct / N, N)
            loss_meter.update(loss.item(), N)
        return {
            "metrics": [
                kddl._logging.loss_metric(loss_meter.avg),
                kddl._logging.Metric(
                    "accuracy", acc_meter.avg, increasing=True
                ),
            ]
        }

    def evaluate_train(self, dl_fn):
        return self.evaluate(dl_fn(self.train_ds()))

    def evaluate_val(self, dl_fn):
        return self.evaluate(dl_fn(self.val_ds()))



"""LR finder that branches from a base training run for a range of learning 
rates.

The training process is:
    1. Train a model with an initial learning rate fully. Fully here means until
    the provided number of steps is reached, or the provided early stopping is
    triggered. 
        - Training is done with a constant learning rate.
        - There are regularly spaced checkpoints (future feature: evaluate
            using the validation dataset).
    2. At regular intervals of the checkpoints, for a short number of steps,
    train the model with a range of learning rates.
    3. Record all loss "branches" and return this data.

The choice is made by taking in this data and choosing the highest learning
rate that at some point was the most beneficial.
"""
import kddl
import kddl.train as train
import polars as pl
import math
import logging
from typing import Tuple, Optional, Sequence, List
import numpy as np
from numpy.lib.stride_tricks import sliding_window_view
import scipy.stats
import itertools

_logger = logging.getLogger(__name__)





"""
Test for test code.

These tests cover the reference implementation used in the pytest code.

The reference implementations are expected to not be modified, and their failure
means there is possibly something wrong with the training infrastructure, or
an accidental change in the reference implementation.

Baseline 1
==========
A CNN for MNIST classification.
"""

import pytest
import torch
import kddl.train
import kddl._logging

# Note on importing nn.py. pytest will add kddl/test to the sys.path, and so,
# any modules in the kddl/test directory can be imported directly.
import kddl.testing as test_nn


def _val_dl_fn(ds):
    val_dl = torch.utils.data.DataLoader(
        ds,
        batch_size=256,
        shuffle=False,
        num_workers=4,
        drop_last=False,
    )
    return val_dl


_default_train_args = {
    "n_epochs": 14,
    "batch_size": 64,
    "lr": 0.001,
    "weight_decay": 0.01,
    "save_checkpoints": False,
}


def test_cnn1(out_root, mnist_ds_mgr, seed_random):
    """Make sure our baseline CNN performs as expected on MNIST.

    TODO: use the best-loss checkpoint for the evaluation.
    TODO: this model isn't actually that reliable! Need to have a reliable
    base model.
    """
    model = test_nn.Cnn1()
    trainable = test_nn.ClassifyTrainable(mnist_ds_mgr, model, "cnn1_mnist")
    out_dir = kddl._logging.get_outdir(out_root, ["train"])
    kddl.train.train(
        trainable,
        out_dir=out_dir,
        **_default_train_args,
    )

    trainable.model.eval()
    metrics = trainable.evaluate_val(_val_dl_fn)
    loss, acc = metrics["metrics"]
    assert acc.value > 0.70, "This should really not fail"
    assert acc.value == pytest.approx(0.850, abs=0.10)


def test_fastai_cnn1(out_root, fashion_mnist_ds_mgr, seed_random):
    model = test_nn.FastaiCnn1()
    trainable = test_nn.ClassifyTrainable(
        fashion_mnist_ds_mgr, model, "fastai_cnn1_fashion_mnist"
    )
    out_dir = kddl._logging.get_outdir(out_root, ["train"])
    kddl.train.train(
        trainable,
        out_dir=out_dir,
        **_default_train_args,
    )
    trainable.model.eval()
    metrics = trainable.evaluate_val(_val_dl_fn)
    loss, acc = metrics["metrics"]
    # assert acc.value > 0.70, "This should really not fail"
    # assert acc.value == pytest.approx(0.850, abs=0.10)








import pytest
import numpy as np
import kddl
import kddl.lrfind as lrfind
from pathlib import Path
import json
import polars as pl
import logging

_logger = logging.getLogger(__name__)


def load_json(case_num):
    path = Path(__file__).parent / f"resources/case{case_num}.json"
    with open(path, "r") as f:
        return json.load(f)


def load_parquet(case_num):
    path = Path(__file__).parent / f"resources/case{case_num}.parquet"
    return pl.read_parquet(path)


def test_cut_takeoff1():
    """
    Tests that cut_takeoff cuts off the right-side of the array.

    The array is a 1D array something like:

                                            
                                            *
                                            *
       * * * * * *                         *
                   *                      *
                     *                    *
                       *                 *
                        * * * * * * * * *
    """
    loss_arr = lrfind.to_loss_arr(load_json(1))
    res = lrfind.cut_takeoff(loss_arr)
    very_low = 766
    very_high = 900
    expected = 800
    _len = res.shape[1]
    assert very_low < _len < very_high, "This should definitely not fail."
    assert _len == pytest.approx(expected, abs=20), f"should be ~{expected}."


def test_cut_takeoff2():
    """
    This test was written in response to a bug. The bug was fixed by changing
    the line:

        `above_max = np.flatnonzero(mask_in * loss_arr.mean(axis=0) > ignore_over)`

    to:

        `above_max = np.flatnonzero(mask_in * (loss_arr.mean(axis=0) > ignore_over))`

    The error case occured when mask_in was False, but ignore_over was below
    zero.
    """
    cutoff_case2 = lrfind.to_loss_arr(load_json(3))
    assert len(cutoff_case2[0]) == 1000
    res = lrfind.cut_takeoff(cutoff_case2)
    assert res.shape[1] > 400, "Shouldn't be severely cut off."


def test_cut_takeoff3():
    """
    This test was witten in response to a perceived probable bug associated with
    cutting off the right side too close to the minimum value. This led to an
    issue when the minimum value was due to a large deviation from the loss on
    either side. Cutting off later (further right) will hopefully prevent this
    spurious low loss being identified as the location of best lr.
    """
    cutoff_case3 = lrfind.to_loss_arr(load_json(6))
    assert len(cutoff_case3[0]) == 1000
    res = lrfind.cut_takeoff(cutoff_case3)
    # 680 is the value where it was cut off before. Probabbly at least 750 is
    # a good value.
    # Okay to tweak:
    assert res.shape[1] > 750, "Shouldn't be cut off so early."
    # Most likely a bug if this fails:
    assert res.shape[1] > 680, "Shouldn't be cut off so early."


def test_cut_takeoff4():
    """This test was written when a case was encountered where cut_takeoff
    returned an empty array. The cause was the loss curve was practically flat,
    and the minimum value happened to be at the start of the array. The correct
    behaviour when the loss curve is flat is to return the whole array.
    Currently, this is only done when cut_takeoff would return an empty array;
    however, it should probably be extended to indentify flat losses with a
    statistical test, which would allow us to detect the case even when it's
    not the very first element that is the minimum value."""
    loss_arr = lrfind.to_loss_arr(load_json(9))

    # Test 1.
    # This is what cut_takeoff uses.
    res = lrfind.is_flat(loss_arr)
    assert res, "Should be flat."

    # Test 2.
    # cut_takeoff doesn't cut off anything.
    res = lrfind.cut_takeoff(loss_arr)
    assert res.shape[1] == loss_arr.shape[1], "Souldn't be any cut off."


def test_cut_takeoff5():
    """Insures the instability and flicks at the end are cut off.

    This test uses case 11.
    """
    df = load_parquet(11)

    # Test 1
    # All sequences should have at least 50 steps removed.
    for i, row in enumerate(df.iter_rows(named=True)):
        loss_arr = lrfind.to_loss_arr(row["loss"])
        res = lrfind.cut_takeoff(loss_arr)
        n_cut = loss_arr.shape[1] - res.shape[1]
        assert (
            n_cut >= 50
        ), f"Should have cut at least 50 steps. ({i=})"


def test_cut_takeoff6():
    """The opposite of test_cut_takeoff5: don't cut off too soon."""
    df = load_parquet(12)

    # Test 1
    # All sequences should have at least 750 steps remaining.
    for i, row in enumerate(df.iter_rows(named=True)):
        loss_arr = lrfind.to_loss_arr(row["loss"])
        res = lrfind.cut_takeoff(loss_arr)
        assert res.shape[1] >= 750, f"Should have at least 750 steps. ({i=})"

def test_is_flat():
    """Insure that is_flat() returns true for a flat case.

    is_flat() previously returned true on this case (an omi-const sweep).
    """
    # Setup
    sweep = np.array(load_json(13))
    N, L = sweep.shape
    assert N == 8 and L == 1000, "Sweep had 8 runs of 1000 steps."

    # Test
    # assert lrfind.is_flat(sweep), "Should be flat."
    assert not lrfind.is_flat(sweep), (
        "This is a known broken case (unexpectedly passed!)"
    )




def test_kalman_smooth():
    """
    The lr curve looks something like:


                                                        *           *
                                                      *  *         *
                                                    *     *       *
                                                  *        *     *
        * * * * * *            * * * * * * * * * *          *   *
                    *         *                              * *
                      *      *                                *
                        *   *
                          *

    We want to choose the first dip.
    """
    loss_arr = lrfind.to_loss_arr(load_json(2))
    smooth_dloss = kddl.lrfind.kalman_smooth(loss_arr)
    min_grad_at = np.argmin(smooth_dloss)
    buf = 50
    assert min_grad_at < len(smooth_dloss) - buf


def test_kalman_2_dips():
    """
    The lr curve looks something like:

        * * * * * *        * * * * * * * * * * * * * * *        *
                   *      *                             *       *
                    *    *                               *     *
                     *  *                                 *   *
                      *                                    * *
                                                            *
    We want to choose the second dip.

    On a figure, the first dip looks visually appealing as a choice for the
    best learning rate; however, there isn't really any reason not to choose
    the second dip, as it doesn't have any pathalogical properties.

    This test case is to ensure that the algorithm chooses the second dip.
    If this tests breaks, then maybe you have either:

      - tried to fix a perceived bug that wasn't a bug, or
      - you have identified why the second dip is a bad choice.

    In the second case, the reasoning should be made clear in an updated
    description of the choice function.
    """
    loss_arr = lrfind.to_loss_arr(load_json(4))
    smooth_dloss = kddl.lrfind.kalman_smooth(loss_arr)
    min_grad_at = np.argmin(smooth_dloss)
    # If this fails, please read the test description before updating the
    # target value. It's reasonable for the value to change slightly, but
    # not jump back to a low value.
    assert min_grad_at == pytest.approx(870, abs=10), "should be near 870."
    # This assert; however, should not be tweaked.
    assert min_grad_at > 500, "Should never fail; function is likely broken."


@pytest.mark.skip(reason="Using v1, which chooses the second dip.")
def test_kalman_2_dips_v2():
    """
    On closer inspection of the lr curve (compare to the disabled test above),
    the lr curve is more like:
                                                                      *
                                                               *     *
        * * * * * *        * * * * * * * * * * * * * *         *    *
                   *      *                            *       *   *
                    *    *                              *     * * *
                     *  *                                *    * **
                      *                                    * *  *
                                                            *   *

    And so there is arguably 3 dips. If we reduce the process variance, then
    the first dip is preferred, as the latter two are smoothed out. From other
    tests, the process variance scaled at sqrt(2) seemed to high, and so it was
    reduced. The algorithm now chooses the first dip. The choice of the first
    or second dip will switch based on the value of Q_scale, the parameter to
    kalman_smooth. When Q_scale is low (tested at 1/3), then the process
    variance is pushed lower, and the second dip is not steady enough and is
    smoothed out. When Q_scale is higher (tested at sqrt(2)), the second dip is
    chosen.

    On a figure, the first dip looks visually appealing as a choice for the
    best learning rate; however, there isn't really any reason not to choose
    the second dip, as it doesn't have any pathalogical properties.

    This test case is to ensure that the algorithm chooses the second dip.
    If this tests breaks, then maybe you have either:

      - tried to fix a perceived bug that wasn't a bug, or
      - you have identified why the second dip is a bad choice.

    In the second case, the reasoning should be made clear in an updated
    description of the choice function.
    """
    loss_arr = lrfind.to_loss_arr(load_json(4))
    smooth_dloss = kddl.lrfind.kalman_smooth(loss_arr)
    min_grad_at = np.argmin(smooth_dloss)
    # If this fails, please read the test description before updating the
    # target value. It's reasonable for the value to change slightly, but
    # not jump back to a low value.
    # History: 461, 470
    assert min_grad_at == pytest.approx(470, abs=10), "should be near 470."
    # This assert; however, should not be tweaked.
    assert min_grad_at < 700, "Should never fail; function is likely broken."


def test_kalman_3():
    """
    The lr curve looks something like:

        * * * * * * **
                       **
                         **
                           **
                            **
                               **            **
                                  ** * * * *
              ↑           ↑
            error       expected

    An error case was observed where a very low learning rate was chosen on
    a nice and simple looking lr curve where there seemed to be an obvious
    sleep slope where the learning rate should have be chosen. This case
    is the basis for the test case.
    """
    loss_arr = lrfind.to_loss_arr(load_json(5))
    smooth_dloss = kddl.lrfind.kalman_smooth(loss_arr)
    min_grad_at = np.argmin(smooth_dloss)
    # If this fails, please read the test description before updating the
    # target value. It's reasonable for the value to change slightly, but
    # not jump back to a low value.
    assert min_grad_at == pytest.approx(481, abs=20)


def test_kalman_4():
    """
    An error was observed where a model with very high loss was having the
    loss overly smoothed, to the point where it want practically a straght
    line, despite the loss having a nice curve and clear good lr region.

    The cause is believed to be due to incorrect choice of process and
    observation variance.
    """
    loss_arr = lrfind.to_loss_arr(load_json(6))
    smooth_dloss = kddl.lrfind.kalman_smooth(loss_arr)
    min_grad_at = np.argmin(smooth_dloss)
    assert min_grad_at == pytest.approx(650, abs=100)


def test_kalman_5():
    """The boundary was pulling the Kalman smoothed loss down.

    Covers two cases, both similar effect.
    """
    # Don't cut off too early.
    loss_arr = lrfind.to_loss_arr(load_json(4))
    res = lrfind.cut_takeoff(loss_arr)
    # 866 is the value where it was cut off before. More would be better.
    # Okay to tweak upwards.
    assert res.shape[1] >= 866, "Shouldn't be cut off so early."

    # Don't be affected by the boundary.
    case = lrfind.to_loss_arr(load_json(8))
    smooth_dloss = kddl.lrfind.kalman_smooth(case)
    min_grad_at = np.argmin(smooth_dloss)
    # Some history
    # This test used to have:
    #
    ## Should be lower than 750, maybe around 600.
    # assert min_grad_at < 700
    ## The cutoff is at 904, so it should definitely be lower than that.
    # assert min_grad_at < 904
    #
    # But on inspecting the lr-curve, it seems this is when we were more keen
    # to avoid placing the best lr in the 2nd dip. The more recent thinking is
    # that the second dip is exactly where we want to place the lr, as we want
    # the just-before-unstable region, and not just target some region of
    # long negative slope. This is based on the updated understanding that the
    # slopes are quite dependent on the dataset, and so we shouldn't pay
    # too much attention to them, or to try and value long negative slopes.
    #
    # I'm not even sure if this test is necessary, or at least its purpose
    # has changed. I'll allow either of the slopes. It's hard to expect one
    # or the other, as they look very similar steepness.
    at_first = 420 < min_grad_at < 550
    at_second = 850 < min_grad_at < 930
    assert at_first or at_second, "Should be in the first or second dip."


def test_kalman_6():
    """Insures the Kalman filter doesn't have a flick at the end.

    This flick has been observed many times. The test data consists of ~50
    cases of the flick that were found in a single experiment.

    The case data is a parquet file, as we need to do two things:
        - make sure we can identify all the flicks in the historic predictions
        - then make sure current predictions don't have the flick.
    """
    df = load_parquet(10)
    # Case 10 also has cases where there is no flick, which we use to ensure
    # the identification of the flick is working.
    df_no_flick = pl.read_parquet(
        Path(__file__).parent / f"resources/case10_noflick.parquet"
    )

    def has_flick(kalman_smoothed):
        TAIL_LEN = min(50, len(kalman_smoothed) // 2)
        tail_slope = (
            abs(kalman_smoothed[-TAIL_LEN - 1] - kalman_smoothed[-1]) / TAIL_LEN
        )
        # Larger q makes it less likely to register a flick.
        # This value is tweaked to reduce false positives and negatives.
        q = 0.60
        slope_threshold = np.quantile(np.abs(np.diff(kalman_smoothed)), q)
        has_flick = tail_slope > slope_threshold
        _logger.info(f"{tail_slope=}, {slope_threshold=}, {has_flick=}")
        return has_flick

    # Test 0: internal test.
    def test_has_flick():
        """Test of the test code, has_flick."""
        for i, row in enumerate(df.iter_rows(named=True)):
            kalman_smoothed = row["ckalman"]
            assert has_flick(kalman_smoothed), (
                "Test is broken. Flicks should be present in this example: "
                f'{i}: {row["ds"]=}, {row["batch_size"]=}, {row["model"]=}'
            )

        for i, row in enumerate(df_no_flick.iter_rows(named=True)):
            kalman_smoothed = row["ckalman"]
            assert not has_flick(kalman_smoothed), (
                "Test is broken. Flicks should be absent in this example: "
                f'{i}: {row["ds"]=}, {row["batch_size"]=}, {row["model"]=}'
            )

    test_has_flick()

    # Test 1: current code produces no flicks.
    df = pl.concat([df, df_no_flick])
    for i, row in enumerate(df.iter_rows(named=True)):
        loss_arr = lrfind.to_loss_arr(row["loss"])
        kalman_smoothed = lrfind.kalman_smooth(loss_arr).tolist()
        has_fl = has_flick(kalman_smoothed)
        if i in {2,4, 39, 46}:
            assert has_fl, (
                "This is a known broken test, and now unexpectedly passed!"
            )
            continue
        assert not has_fl, (
            "Flicks should be absent. "
            f'{i=}: {row["ds"]=}, {row["batch_size"]=}, {row["model"]=}'
        )



import pytest
import kddl._logging as lgg
import time
import logging
import polars as pl
import polars.testing


def test_Timer():
    """Tests a Timer to insure that:

    1. stop() can be called without start()
    2. calling any of elapsed(), total_elapsed() or rolling_duration()
       before start() raises an exception.
    3. rolling duration is incremented correctly:
       3.1. after create_and_start() it is 0
       3.2. after one call it is the elapsed time
       3.3. after two calls it is (1-w)(first_call) + w(second_call)

    """
    # 1. stop() can be called without start()
    t = lgg.Timer()
    t.stop()

    # 2. before start() exceptions.
    # 2.1 elapsed()
    t = lgg.Timer()
    with pytest.raises(ValueError):
        t.elapsed()
    # 2.2 total_elapsed()
    t = lgg.Timer()
    with pytest.raises(ValueError):
        t.total_elapsed()
    # 2.3 rolling_duration()
    t = lgg.Timer()
    with pytest.raises(ValueError):
        t.rolling_duration()

    # 3. rolling duration
    # 3.1 after create_and_start() it is 0
    t = lgg.Timer.create_and_start()
    assert t.rolling_duration() == pytest.approx(0, abs=1e-6)

    # 3.2 after one call it is the elapsed time
    t = lgg.Timer.create_and_start()
    time.sleep(0.5)
    t.restart()
    assert t.rolling_duration() == pytest.approx(0.5, abs=1e-2)

    # 3.3 after two calls it is (1-w)(first_call) + w(second_call)
    time.sleep(0.01)
    t.restart()
    assert t.rolling_duration() == pytest.approx(
        0.01 * t.w + 0.5 * (1 - t.w), abs=1e-3
    )


def test_get_outdir(tmp_path_factory, caplog):
    """
    Tests that:
       1. Missing directories are skipped when picking next version number.
       2. If base directory doesn't exist, it is created.
       3. A warning is issued when creating a version value >= 1000.
            3.1 When there are actually all versions 0->999.
            3.2 Even when there are hardly any folders, say just one: 999.
    """
    # 1. Missing directories are skipped when picking next version number.
    # Setup
    # Create a sub-directory structure like:
    #   ./exp/1/7/cnn/0
    #   ./exp/1/7/cnn/1
    #   ./exp/1/7/cnn/3
    tmp_path = tmp_path_factory.mktemp("exp")
    tmp_path.joinpath("1/7/cnn/0").mkdir(parents=True)
    tmp_path.joinpath("1/7/cnn/1").mkdir(parents=True)
    tmp_path.joinpath("1/7/cnn/3").mkdir(parents=True)
    # Test
    outdir = lgg.get_outdir(tmp_path, ["1", "7", "cnn"])
    assert outdir == tmp_path.joinpath("1/7/cnn/4")
    assert outdir.exists()

    # 2. If base directory doesn't exist, it is created.
    # Setup
    tmp_path = tmp_path_factory.mktemp("out")
    # Test
    base_dir = tmp_path / "out"
    outdir = lgg.get_outdir(base_dir, ["1", "2", "3"])
    assert base_dir.exists()
    assert base_dir.joinpath("1/2/3/0").exists()

    # 3. A warning is issued when creating a version value > 1000.
    # 3.1 When there are actually all versions 0->999.
    with caplog.at_level(logging.WARNING):
        # Setup
        tmp_path = tmp_path_factory.mktemp("exp3")
        for _ in range(1000):
            lgg.get_outdir(tmp_path, ["1", "7", "cnn"])
        assert not caplog.records
        lgg.get_outdir(tmp_path, ["1", "7", "cnn"])
        # Test
        assert caplog.records, "Reached outdir version 1000."
    # 4.2 When there is a gap until 999.
    with caplog.at_level(logging.WARNING):
        # Setup
        caplog.clear()
        assert not caplog.records, "If this fails, I don't understand caplog."
        tmp_path = tmp_path_factory.mktemp("exp4")
        tmp_path.joinpath("1/7/cnn/999").mkdir(parents=True)
        assert not caplog.records
        # Test
        lgg.get_outdir(tmp_path, ["1", "7", "cnn"])
        assert caplog.records, "Reached outdir version 1000."


def test_MetricTracker(tmp_path):
    """Tests that:
    1. Recording doesn't raise an error and a dataframe can be created.
        1.1 The MetricTracker records by epoch.
        1.1 The MetricTracker records by epoch and step.
    """
    # Setup
    mt1 = lgg.MetricTracker(tmp_path)
    mt2 = lgg.MetricTracker(tmp_path)
    # fmt: off
    metrics = [
        (0, 1, lgg.Metric("loss", 7, increasing=False, checkpointed=True), True),
        (0, 10, lgg.Metric("loss", 8, increasing=False, checkpointed=True), False),
        (0, 100, lgg.Metric("loss", 6, increasing=False, checkpointed=True), True),
        (1, 1, lgg.Metric("loss", 5.2, increasing=False, checkpointed=True), True),
        (1, 10, lgg.Metric("loss", 6.2, increasing=False, checkpointed=True), False),
    ]
    # fmt: on
    expected_df1 = pl.DataFrame(
        [[0, None, 7], [1, None, 5.2]],
        schema=pl.Schema(
            {"epoch": pl.UInt32, "step": pl.UInt64, "loss": pl.Float64}
        ),
        orient="row",
    )
    expected_df2 = pl.DataFrame(
        [[0, 1, 7], [0, 10, 8], [0, 100, 6], [1, 1, 5.2], [1, 10, 6.2]],
        schema=pl.Schema(
            {"epoch": pl.UInt32, "step": pl.UInt64, "loss": pl.Float64}
        ),
        orient="row",
    )

    # Test
    # 1.
    prev_epoch = -1
    for epoch, step, metric, is_better in metrics:
        if epoch != prev_epoch:
            new_best1 = mt1.on_epoch_end([metric], epoch)
            prev_epoch = epoch
            if is_better:
                assert len(new_best1)
        new_best2 = mt2.new_metrics([metric], epoch, step)
        if is_better:
            assert len(new_best2)

    df1 = mt1.history_as_dataframe()
    df2 = mt2.history_as_dataframe()
    pl.testing.assert_frame_equal(df1, expected_df1)
    pl.testing.assert_frame_equal(df2, expected_df2)



import logging
import math
from typing import Callable, Optional, Union
from numpy.typing import ArrayLike
import einops
import torch
import torch.nn as nn
from torch.nn import functional as F


_logger = logging.getLogger(__name__)


def get_optim_param_groups(module):
        """Return parameters with and without decay, in separate groups.

        This function signature is designated by train.py.
        """
        with_grad = {
            n: p for n, p in module.named_parameters() if p.requires_grad
        }
        # Collect set of all embedding parameters.
        embedding_params = set()
        for module in module.modules():
            if isinstance(module, nn.Embedding):
                for param in module.parameters(recurse=False):
                    embedding_params.add(id(param))

        def with_decay_fn(param):
            single_dim = param.dim() == 1
            is_embedding = id(param) in embedding_params
            with_decay = not (single_dim or is_embedding)
            return with_decay

        with_decay_params = []
        without_decay_params = []
        with_decay_names = []
        without_decay_names = []
        for n, p in with_grad.items():
            if with_decay_fn(p):
                with_decay_params.append(p)
                with_decay_names.append(n)
            else:
                without_decay_params.append(p)
                without_decay_names.append(n)
        _logger.info(f"Params with decay: {', '.join(with_decay_names)}")
        _logger.info(f"Params without decay: {', '.join(without_decay_names)}")
        return [with_decay_params, without_decay_params]


class NormMask(nn.Module):
    """Normalize, but set masked values to 0.

    Sequences can contain padding, which uses some padding value (e.g. 0, -1).
    We want to normalize these sequences, which will send padded values to
    some arbitrary value. To handle this, we set the masked values to 0 after
    normalization. We have lost the original distinction between padding and
    real values—this is the cost of normalization.

    Given that the mask will be needed (otherwise one would just use 
    Normalize), we will go ahead and concatenate the mask to the input. This
    implementation assumes a channel-last format.

    We could alternatively take in the pad value and construct and return the
    mask here. So for this hasn't been needed as a mask has been available as
    it is used by Trainables.
    """
    def __init__(self):
        super().__init__()
        self.register_buffer("mean", torch.zeros(size=[]))
        self.register_buffer("sd", torch.ones(size=[]))

    def set_mean_sd(self, m: ArrayLike, sd: ArrayLike):
        self.mean = torch.tensor(m)
        self.sd = torch.tensor(sd)
        if self.mean.ndim != 0:
            raise ValueError(
                f"Expected mean to be a scalar; got ({self.mean.ndim}) dims."
            )
        if self.sd.ndim != 0:
            raise ValueError(
                f"Expected sd to be a scalar; got ({self.sd.ndim}) dims."
            )

    def forward(self, x, mask):
        """
        Args:
            x [batch, seq_len, n_features]
            mask [batch, seq_len]:  1s for valid, 0s for invalid.
        Returns:
            x [batch, seq_len, n_features+1]
        """
        x = self.norm(x)
        mask = einops.rearrange(mask, "b s -> b s 1")
        x = x * mask
        x = torch.cat([x, mask], dim=-1)
        return x

    def norm(self, x):
        return (x - self.mean) / self.sd

    def denorm(self, x):
        return x * self.sd + self.mean


class Normalize(nn.Module):
    def __init__(self):
        super().__init__()
        self.register_buffer("mean", torch.zeros(size=[]))
        self.register_buffer("sd", torch.ones(size=[]))

    def set_mean_sd(self, m: ArrayLike, sd: ArrayLike):
        self.mean = torch.tensor(m)
        self.sd = torch.tensor(sd)
        if self.mean.ndim != 0:
            raise ValueError(
                f"Expected mean to be a scalar; got ({self.mean.ndim}) dims."
            )
        if self.sd.ndim != 0:
            raise ValueError(
                f"Expected sd to be a scalar; got ({self.sd.ndim}) dims."
            )

    def forward(self, x, shift=True, scale=True):
        shift = self.mean if shift else 0
        scale = self.sd if scale else 1
        return (x - shift) / self.sd

    def denorm(self, x):
        return x * self.sd + self.mean


class IdxEmbed(nn.Module):
    def __init__(self, length, n_embd):
        super().__init__()
        self.n_embd = n_embd
        self.length = length
        self.pos_embd = nn.Embedding(self.length, self.n_embd)
        self.register_buffer(
            "range_cache",
            einops.rearrange(torch.arange(self.length), "L -> 1 L"),
            persistent=False,
        )

    def forward(self):
        return self.pos_embd(self.range_cache)


class ValueEmbed(nn.Module):
    """Convert scalars to a fixed sized vector.

    Can be used for timestamps or single dimension data.
    """

    def __init__(self, n_embd):
        super().__init__()
        self.n_embd = n_embd
        self.register_buffer("slope", torch.zeros(self.n_embd // 2))

    @torch.no_grad()
    def init_weights(self, max_range, epsilon):
        # Use up most of the range [0, 2π] by landing on 3/4 * 2π.
        min_scale = (3 / 2 * math.pi) / max_range
        max_scale = (3 / 2 * math.pi) / epsilon
        self.slope.data = torch.exp(
            torch.linspace(
                math.log(min_scale),
                math.log(max_scale),
                self.n_embd // 2,
                dtype=torch.float32,
                requires_grad=False,
            )
        )

    def forward(self, t):
        t = self.slope * t#.unsqueeze(-1)
        t = torch.cat([t.sin(), t.cos()], dim=-1).to(dtype=torch.float32)
        return t


def create_batch_norm(n: int) -> nn.Module:
    """
    Following Davis and Frank's recommendation in "Revisiting Batch
    Normalization", we would initialize batch norm weights to less than 1
    (they suggested to use 0.1). They also recommended using a lower learning
    rate for the γ parameter.

    For comparison, fastai initialize β to 0.001 and γ to 1.

    I tried both, and found better results with fastai's defaults.
    """
    bn = nn.BatchNorm1d(n)
    # fastai
    # bn.weight.data.fill_(1.0)
    # bn.bias.data.fill_(1e-3)
    # Davis and Frank
    bn.weight.data.fill_(0.1)
    bn.bias.data.fill_(0)
    return bn


def create_shortcut(in_n, out_n, stride, num_dim=1):
    """Residual connection.

    The identify path is one of those finicky bits of ResNet type networks.

    Depending on whether the input and output match in terms of channel
    count and dimension, we have the following behaviour:

    Match?
    ------

        | Channel count | Dimensions | Bevaviour                  |
        |---------------|------------|----------------------------|
        |      ✓        |     ✓      | identity                   |
        |      ✓        |     ✘      | pool or conv               |
        |      ✘        |     ✓      | 1x1 conv                   |
        |      ✘        |     ✘      | pool or conv and 1x1 conv  |


    The most interesting choice is whether to use a strided pool or a
    strided convolution to achieve the downsampling effect. It's
    interesting as implementations are split on which one to use. There
    are further choices too, such as whether to use dilation in addition
    to strides, and whether to downsample before or after the 1x1 conv.

    Some implementations for reference:
        - fastai: https://github.com/fastai/fastai/blob/aa58b1316ad8e7a5fa2e434e15e5fe6df4f4db56/nbs/01_layers.ipynb
        - lightning: https://github.com/Lightning-AI/lightning-bolts/blob/master/pl_bolts/models/self_supervised/resnets.py
        - pytorch image models: https://github.com/rwightman/pytorch-image-models/blob/master/timm/models/resnet.py
             - there are two functions, downsample_conv() and
               downsample_avg() that are used to create the downsampling for
               the shortcut connection.
             - uses ceil_mode=True, so a dimension n=7 would be reduced to
               n=4, not n=3.

    My gut instinct is that swapping out pooling for convolution to achieve
    the downsample will naively achieve better results; however, the
    convolution is intrinsically more powerful (actual has parameters) and
    so, if you care about parameter counts, then a fair comparison would
    involve reducing parameters elsewhere. Given that the whole point of
    the residual layer is to short circuit a layer and allow gradients to
    flow easily, I think that pooling gets more theory points for staying
    true to this idea. Ironically, I think the original ResNet
    implementation might have used convolution."""
    # 1. Identity
    # In the simplest case, we can just return the input. For this to work,
    # both the channel count and channel dimensions of the input and output
    # must match. The output channel dimension is determined by the stride.
    if num_dim == 1:
        poolXdClass = nn.AvgPool1d
        convXdClass = nn.Conv1d
    elif num_dim == 2:
        poolXdClass = nn.AvgPool2d
        convXdClass = nn.Conv2d
    else:
        raise ValueError(f"Only 1d and 2d supported; got ({num_dim}).")
    channels_match = in_n == out_n
    downsample = stride > 1
    if channels_match and not downsample:
        return nn.Identity()

    skip_layers = []
    # 2. Downsample
    if downsample:
        # The following (1, 7) input:
        # |  1  |  2  |  3  |  4  |  5  |  6  |  7  |
        # when pooled with stride=kernel=2 becomes (1, 4):
        # |    1.5    |    3.5    |    5.5   | 7 |
        pool = poolXdClass(
            kernel_size=stride,
            stride=stride,
            count_include_pad=False,
            ceil_mode=True,
        )
        skip_layers.append(pool)
    # 3. 1x1 conv
    if not channels_match:
        # There isn't a consensus on whether:
        #   - to use batch norm or a bias or neither.
        conv = convXdClass(in_n, out_n, kernel_size=1, bias=False)
        # TODO: how is best to initialize this connection?
        skip_layers.append(conv)
    res = nn.Sequential(*skip_layers)
    return res


@torch.no_grad()
def sinusoidal_embedding(length: int, n_ch: int) -> torch.Tensor:
    """
    Create a tensor of shape [time_steps, n_channels//2] and fill all rows with
    the sequence: [0, 1, 2, ..., length-1]. Then, scale each row by a different
    value. The list of scaling values are chosen to span from 1 to 1/10000.
    The 0th channel is scaled by 1, then gradually the scale is reduced
    until the last channel, (n_ch//2 -1), is scaled by 1/10000. The steps are
    exponentially spaced, so the scaling will initially rapidly decrease, then
    slow down as the minimum is approached.
    This resulting array is used as input to two functions: sine and cos, and
    the two results are concatenated in order to get a tensor of shape
    [time_steps, n_channels].
    """
    half_nch = n_ch // 2
    # From 1 to 1/10000, exponentially in half_nch steps.
    slope = (
        torch.arange(half_nch, dtype=torch.float)
        * -math.log(10000)
        / (half_nch - 1)
    ).exp()
    t = slope[:, None] * torch.arange(length, dtype=torch.float)[None, :]
    res = torch.cat([t.sin(), t.cos()], dim=0).to(dtype=torch.float32)
    return res


class Attention(nn.Module):
    def __init__(
        self, n_embd, heads=4, dropout=0.0, project_out=True, is_causal=False
    ):
        super().__init__()
        self.project_out = project_out
        self.dim_head = int(n_embd / heads)
        if self.dim_head != n_embd / heads:
            raise ValueError(
                f"n_embd ({n_embd}) must be divisible by heads ({heads})"
            )
        self.heads = heads
        self.is_causal = is_causal

        self.to_qkv = nn.Linear(n_embd, n_embd * 3, bias=False)
        self.to_out = (
            nn.Sequential(
                nn.Linear(n_embd, n_embd),
                nn.Dropout(dropout),
            )
            if self.project_out
            else nn.Identity()
        )
        self.norm = nn.LayerNorm(n_embd)
        self.dropout_rate = dropout

    def forward(self, x, mask=None):
        x = self.norm(x)
        qkv = self.to_qkv(x).chunk(3, dim=-1)
        q, k, v = map(
            lambda t: einops.rearrange(t, "b t (h c) -> b h t c", h=self.heads),
            qkv,
        )

        # q = q * self.scale
        # raw_attn = torch.einsum("b h i d, b h j d -> b h i j", q, k)
        # attn = raw_attn.softmax(dim=-1)
        # attn = self.dropout(attn)
        # out = torch.einsum("b h i j, b h j d -> b h i d", attn, v)
        dropout_rate = self.dropout_rate if self.training else 0.0
        # Note: can't set both attN_mask and is_causal, so do our own masking.
        out = F.scaled_dot_product_attention(
            q, k, v, dropout_p=dropout_rate, is_causal=self.is_causal
        )
        out = einops.rearrange(out, "b h t d -> b t (h d)")
        out = self.to_out(out)
        if mask is not None:
            out = out * mask
        return out


class FeedForward(nn.Module):
    def __init__(self, n_embd, hidden_dim, dropout=0.0):
        super().__init__()
        self.input_norm = nn.LayerNorm(n_embd)
        self.linear1 = nn.Linear(n_embd, hidden_dim)
        self.dropout = nn.Dropout(dropout)
        self.linear2 = nn.Linear(hidden_dim, n_embd)

    def forward(self, x):
        x = self.input_norm(x)
        x = self.linear1(x)
        x = self.dropout(F.gelu(x))
        x = self.linear2(x)
        x = self.dropout(x)
        return x


class Transformer(nn.Module):
    def __init__(
        self, dim, depth, heads, mlp_dim, dropout=0.0, is_causal=False
    ):
        super().__init__()
        self.layers = nn.ModuleList([])
        for _ in range(depth):
            self.layers.append(
                nn.ModuleList(
                    [
                        Attention(
                            dim,
                            heads,
                            dropout,
                            project_out=True,
                            is_causal=is_causal,
                        ),
                        FeedForward(dim, mlp_dim, dropout=dropout),
                    ]
                )
            )

    def forward(self, x, attn_mask=None):
        # TODO: rework mask handling, as it's not really the classic attn mask.
        if attn_mask is not None:
            attn_mask = einops.rearrange(attn_mask, "b s -> b s 1")
        for attn, ff in self.layers:
            x = attn(x, attn_mask) + x
            x = ff(x) + x
        return x


class SEModule(nn.Module):
    def __init__(self, n_channels, reduction=1):
        super().__init__()
        n_mid = n_channels // reduction
        self.fc1 = nn.Conv1d(n_channels, n_mid, kernel_size=1)
        self.relu = nn.ReLU(inplace=True)
        self.fc2 = nn.Conv1d(n_mid, n_channels, kernel_size=1)
        self.sigmoid = nn.Sigmoid()

    def init_weights(self):
        nn.init.kaiming_normal_(self.fc1.weight, mode="fan_in")
        nn.init.xavier_uniform_(self.fc2.weight)

    def forward(self, x):
        module_input = x
        x = x.mean((2,), keepdim=True)
        x = self.fc1(x)
        x = self.relu(x)
        x = self.fc2(x)
        x = self.sigmoid(x)
        return module_input * x

class ResBlock1d(nn.Module):
    """A residual block with 1d convolutions.

    This is a pretty inflexible implementation. No need to make it any
    more general yet.
    """

    def __init__(
        self,
        in_n,
        mid_n,
        out_n,
        kernel_size=3,
        downsample=False,
        dropout: float = 0,
        act_fn: Optional[Callable[[torch.Tensor], torch.Tensor]] = F.relu,
        norm_factory: Optional[
            Callable[[int], torch.nn.Module]
        ] = create_batch_norm,
    ):
        """
        Args:
            act_fn: A function like F.relu.
            norm_factory: a function that returns a Pytorch module. Other
                libraries take a class, and call Class(...) making the __init__
                effectively a factory. However, if you want to do any parameter
                initialization or anything, the interface must be a standard
                function.
        """
        super(ResBlock1d, self).__init__()
        self.downsample = downsample
        self.dropout_rate = dropout
        self.act_fn = identity_fn if act_fn is None else act_fn
        self.norm_factory = (
            create_identity if norm_factory is None else norm_factory
        )
        stride = 2 if self.downsample else 1
        self.shortcut = create_shortcut(in_n, out_n, stride=stride)
        # Note: bias is False for the conv layers, as they will be followed
        # by batch norm.
        self.conv1 = nn.Conv1d(
            in_n,
            mid_n,
            kernel_size=1,
            stride=1,
            padding=0,
            dilation=1,
            bias=False,
        )
        padding = (kernel_size - 1) // 2
        self.conv2 = nn.Conv1d(
            mid_n,
            mid_n,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            dilation=1,
            bias=False,
        )
        self.conv3 = nn.Conv1d(
            mid_n,
            out_n,
            kernel_size=1,
            stride=1,
            padding=0,
            dilation=1,
            bias=False,
        )
        self.dropout = (
            nn.Dropout(p=self.dropout_rate)
            if self.dropout_rate > 0
            else nn.Identity()
        )
        # To use batch norm or group norm?
        self.bn1 = self.norm_factory(mid_n)
        self.bn2 = self.norm_factory(mid_n)
        self.bn3 = self.norm_factory(out_n)

        self.se = SEModule(out_n)

    def init_weights(self):
        nn.init.kaiming_normal_(
            self.conv1.weight, mode="fan_in", nonlinearity="relu"
        )
        nn.init.kaiming_normal_(
            self.conv2.weight, mode="fan_in", nonlinearity="relu"
        )
        if self.act_fn == F.relu:
            nn.init.kaiming_normal_(
                self.conv3.weight, mode="fan_in", nonlinearity="relu"
            )
        else:
            _logger.info(
                "conv3 has no activation function; no attempt is made "
                "to guess a good normalization."
            )

    def forward(self, x):
        shortcut = self.shortcut(x)
        x = self.act_fn(self.bn1(self.conv1(x)))
        x = self.act_fn(self.bn2(self.conv2(x)))
        x = self.dropout(x)
        x = self.act_fn(self.bn3(self.conv3(x)))
        # todo: old or new?
        # old:
        #x = self.se(x) + shortcut
        #x = self.act_fn(x)
        # new:
        x = x + shortcut
        x = self.se(x)
        return x



import math
import torch
import bisect
import numpy as np
from numpy.typing import ArrayLike
from typing import Sequence, Literal, Tuple
import logging
import itertools
import deprecated

_logger = logging.getLogger(__name__)


"""
Useful methods:

Convert 1D array to a 2D "rolling-window" matrix:
lib.stride_tricks.sliding_window_view(x, window_shape, ...)
"""


def estimate_batch_size(sample):
    """Estimate the batch size of a sample."""
    if isinstance(sample, (list, tuple)):
        return len(sample[0])
    elif isinstance(sample, dict):
        return len(next(iter(sample.values())))
    else:
        raise ValueError(f"Unknown sample type. {type(sample)}")


def constrained_iter(iter, max_n=None, len_fn=estimate_batch_size):
    """Enumerate a batched iterable up to max_n samples."""

    max_n = max_n or math.inf
    n = 0
    for sample in iter:
        n += len_fn(sample)
        yield sample
        if n >= max_n:
            break


class ConstrainedIterable:
    """An iteratable limited by number of samples, which are batched.

    The number of samples can be at most max_n + batch_size - 1.
    """

    def __init__(self, iterable, max_n, len_fn=estimate_batch_size):
        self.iterable = iterable
        self.max_n = max_n
        self.len_fn = len_fn

    def __iter__(self):
        return constrained_iter(iter(self.iterable), self.max_n, self.len_fn)

    def __getattr__(self, name):
        """Pass other function calls to the underlying iterable."""
        return getattr(self.iterable, name)


def split_seq1d(seq: ArrayLike, ratios: ArrayLike, interval_len: int):
    """Split an index sequence by splitting [0, t_max] by a given ratio.

    A sequence like [3, 10, 15, 16] that points into a time interval [0, 20]
    would be split into [[3], [10], [15, 16]] with ratios [1, 1, 1].

    [0, 1, 2, 3, 4, 5, 6, 7], [8, 9, 10, 11, 12, 13], [14, 15, 16, 17, 18, 19]

    Args:
        seq: sequence of indices, such as [0, 25, 54, 100]
        ratios: list of ratios for splitting the sequence
        interval_len: the length of the time interval
    """
    ratios = np.array(ratios)
    if len(ratios.shape) != 1:
        raise ValueError("ratios must be a 1D array.")
    boundaries = split_borders(ratios, interval_len)
    seqs = np.split(seq, np.searchsorted(seq, boundaries))
    return seqs


def split_seq(seq, ratios, interval_len, key=lambda x: x):
    """Split an index sequence by splitting [0, t_max] by a given ratio.

    The input list is assumed (not checked) to be sorted.

    This generalizes split_seq1d to work with any list of objects.

    Args:
        seq: sequence of indices, such as [0, 25, 54, 100]
        ratios: list of ratios for splitting the sequence
        interval_len: the length of the time interval
        key: function to extract the value to compare for sorting
    """
    borders = split_borders(ratios, interval_len)
    lo = 0
    res = []
    for b in borders:
        i = bisect.bisect_left(seq, b, lo, key=key)
        res.append(seq[lo:i])
        lo = i
    res.append(seq[lo:])
    return res


def split_borders(ratios: ArrayLike, len: int):
    """
    Calculates the inner borders to split an array by the given ratios.

    Example:

        split_borders((5, 3, 2), 10) -> (5, 8)

    This corresponds to the following splits:

        [0, 1, 2, 3, 4], [5, 6, 7], [8, 9]
         0  1  2  3  4    5  6  7    8  9
                          ^          ^

    Note that (5, 3, 2) are ratios, not the array data. The array data is
    not needed and isn't an argument to this function.
    """
    ratios = np.array(ratios) / np.gcd.reduce(ratios)
    n_divisions = np.sum(ratios)
    if n_divisions > len:
        raise ValueError(
            "The sum of ratios must be less than or equal to the "
            f"length. ({n_divisions} > {len})"
        )
    num_per_division, remainder = divmod(len, n_divisions)
    borders = np.cumsum(ratios * num_per_division) + remainder
    assert borders[-1] == len
    res = borders[:-1].astype(int).tolist()
    return res


def split(arr, ratios):
    """
    Splits an array along the first axis by the given ratios.

    Remaining elements are given to the first split.
    """
    splits = np.split(arr, split_borders(ratios, len(arr)))
    return splits


def split2(data: Sequence, split_ratio: Sequence[int]):
    """
    Splits a sequence by the given ratios.

    This is a non-numpy alternative to split() above.
    """
    divisions = sum(split_ratio)
    if divisions > len(data):
        raise ValueError(
            "The sum of ratios must be less than or equal to the length. "
            f"({divisions} > {len})"
        )
    num_per_division, remainder = divmod(len(data), divisions)
    num_per_split = [num_per_division * ratio for ratio in split_ratio]
    # Give the extra samples to the first split
    num_per_split[0] += remainder
    split_starts = itertools.accumulate([0] + num_per_split)
    splits = [data[s:e] for s, e in itertools.pairwise(split_starts)]
    assert sum(len(s) for s in splits) == len(data)
    return splits


def decompress_seq(seq, t_max, downsample_factor=1):
    seq_ds = np.floor_divide(seq, downsample_factor)
    res = np.zeros(shape=[math.ceil(t_max / downsample_factor)], dtype=int)
    # Allow for multiple events at the same time, which can happen either due
    # to the original sequence having concurrent events or due to downsampling
    # placing multiple events in the same bin.
    np.add.at(res, seq_ds, 1)
    return res


# Switch to warnings.deprecated once Python 3.13.
@deprecated.deprecated("Specify samples_per_epoch instead (input to train()).")
class LongerDataset:
    """
    Loops over a dataset to have a given length.

    All other function calls or attribute accesses are passed through to the
    original dataset.

    The remainder of the division is assigned to first elements of the dataset.
    To fix this, there would need to be random sampling that changes on
    each epoch, which is the responsibility of the DataLoader. So, it's not
    quite so easy to address. So we live with this limitation for now.
    """

    def __init__(self, dataset, length):
        self.dataset = dataset
        self.length = length

    @staticmethod
    def repeat(dataset, n_repeats):
        return LongerDataset(dataset, len(dataset) * n_repeats)

    def __len__(self):
        return self.length

    def __getitem__(self, idx):
        return self.dataset[idx % len(self.dataset)]

    def __getattr__(self, name):
        return getattr(self.dataset, name)


def slice_to_cum_len(
    seqs, length, how: Literal["exact", "under", "over"] = "exact"
):
    """Slice a list of sequences to a cumulative length.

    The `how` parameter determines whether the last sequence is cut to match the target
    length, or whether it is dropped (under) or included in full (over).
    """
    if len(seqs) == 0:
        return []
    cum_len = 0
    for i, seq in enumerate(seqs):
        cum_len += len(seq)
        if cum_len >= length:
            break
    assert i is not None
    if cum_len < length:
        raise ValueError(f"Not enough data to reach length {length}.")
    if how == "over":
        res = seqs[: i + 1]
    elif how == "exact":
        res = seqs[: i + 1]
        res[-1] = res[-1][: length - cum_len]
    elif how == "under":
        res = seqs[:i]
    else:
        raise ValueError(f"Invalid value for how: {how}")
    return res


def num_windows(seq_len: int, win_len: int, stride: int) -> int:
    """Returns the number of position a window can be placed along a sequence.

    Assume for sake of explanation that the dimension the window will move
    along is time. The window starts at the first timestep (t=0) and
    moves forward by stride timesteps to the next position. The window will
    not move forward if any part of it would be outside the sequence.

    Args:
        seq_len: length of the sequence. Or can be thought of as the length
            of which ever dimension the window is moving along.
        win_len: length of the window.
        stride: number of timesteps to move forward for each step.
    """
    # The most likely place for insidious bugs to hide is here. So try two
    # ways.
    # 1. Find the last strided timestep, then divide by stride.
    last_nonstrided = seq_len - win_len
    last_strided = last_nonstrided - (last_nonstrided % stride)
    n_windows1 = last_strided // stride + 1
    # 2. Same deal, but mindlessly copying the formula from Pytorch docs.
    # https://pytorch.org/docs/stable/generated/torch.nn.Conv1d.html
    # Padding is 0, and dilation is 1.
    n_windows2 = math.floor((seq_len - (win_len - 1) - 1) / stride + 1)
    assert n_windows1 == n_windows2, (
        f"Strided snippet calculation is wrong. {n_windows1} != "
        f"{n_windows2}."
    )
    return n_windows1


def num_windows_2d(
    seq_lens: "np.ndarray", win_len: int, stride: int
) -> "np.ndarray":
    """
    A 2D version of n_windows() where there is a list of sequences of different
    lengths.

    The implementation is the same! But is array based. The above function is
    therefore not really needed, but we keep it around anyway as sometimes
    you don't want or need to think of multi-dimensional arrays.
    """
    last_nonstrided = seq_lens - win_len
    last_strided_timestep = last_nonstrided - (last_nonstrided % stride)
    n_windows1 = last_strided_timestep // stride + 1
    # 2. Same deal, but mindlessly copying the formula from Pytorch docs.
    # https://pytorch.org/docs/stable/generated/torch.nn.Conv1d.html
    # Padding is 0, and dilation is 1.
    n_windows2 = np.floor((seq_lens - (win_len - 1) - 1) / stride + 1)
    assert np.array_equal(n_windows1, n_windows2), (
        f"Strided snippet calculation is wrong. {n_windows1} != "
        f"{n_windows2}."
    )
    return n_windows1


def decode_strided_wrap(
    index, seq_len, win_len, stride, shuffle
) -> Tuple[int, int]:
    """Decode an index to the (y, x) coordinates of a window in a 2D array.

    A 1D window moves along a dimension of a sequence and when it reaches the
    end, it wraps around and increments another dimension. This function
    converts a scalar index to the (y, x) coordinates of the window in a
    2D array. If shuffle is enabled, then to the y-coordinate a random number
    less than the stride is added to the y-coordinate. The random number is
    chosen uniformly in the range [0, stride), or [0, max) where max represents
    the largest number less than stride such that the window does not overflow
    the sequence length.

    Args:
        index: the index to decode.
        seq_len: the length of the sequence.
        win_len: the length of the window.
        stride: the number of timesteps to move forward for each step.
        shuffle: whether to add a random number to the window coordinate.
    """
    # Last position a window can be placed at:
    last_nonstrided = seq_len - win_len
    # Last position in [0, w, 2w, 3w, ...] that a window can be placed at:
    last_strided = last_nonstrided - (last_nonstrided % stride)
    n_ys = last_strided // stride + 1
    y = stride * (index % n_ys)
    if shuffle and stride > 1:
        y += torch.randint(0, stride, (1,), dtype=torch.long).item()
        # Randomly shift forward, but not past the last legal window position.
        y = min(y, last_nonstrided)
    x = index // n_ys
    return y, x


def decode_strided_ragged_wrap(
    idx, seq_lens, win_len, stride, shuffle
) -> Tuple[int, int]:
    """
    Decodes the index into the timestep and cell id.

    Similar to _decode_strided_wrap, but for the case where each cell has a
    different number of timesteps. The difference in arguments is that
    num_strided_timesteps is a 1D array instead of a single value.

    Args:
        index (int): The index to decode. In the range [0, ds_len)
        stride (int): The stride used to create the dataset.
        num_strided_timesteps (np.ndarray[int]): The number of strided
        timesteps for each cell.
        max_start_t (np.ndarray[int]): The maximum timestep a snippet can
        start, for each cell.

    """
    seq_lens = np.array(seq_lens) # Ensure it's a numpy array
    n_seqs = len(seq_lens)
    # Last positions a window can be placed at:
    last_nonstrided = seq_lens - win_len
    # Last positions in [0, w, 2w, 3w, ...] that a window can be placed at:
    last_strided = last_nonstrided - (last_nonstrided % stride)
    # Number of elments in [0, w, 2w, 3w, ...] for each sequence.
    n_ys = last_strided // stride + 1

    cumulative_starts = np.concatenate([[0], n_ys.cumsum()])
    # right and left bisects only differ when the requested value equals one of
    # the values in the list.
    #   - bisect_left([0, 5], 0) returns 0
    #   - bisect_right([0, 5], 0) returns 1.
    # Don't try the following line:
    ## seq_idx = bisect.bisect_left(self.cumulative_starts, idx)
    # Why? Example case: [0, 5, 12], idx=4 => 1 (correct is 0)
    x = bisect.bisect_right(cumulative_starts, idx) - 1
    if x >= n_seqs:
        raise ValueError(
            f"Index {idx} is out of bounds for the number of sequences "
            f"{n_seqs} ({cumulative_starts=}, {idx=})"
        )
    y = stride * (idx - cumulative_starts[x])
    if shuffle and stride > 1:
        y += torch.randint(0, stride, (1,), dtype=torch.long).item()
        # Randomly shift forward, but not past the last legal window position.
        y = min(y, last_nonstrided[x])
    return y, x
